'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import xml.etree.ElementTree as ETree
import re

from arjuna.lib.utils import sys_utils
from arjuna.lib.utils import etree_utils
from arjuna.unitee.utils import run_conf_utils
from arjuna.unitee.markup import mrules

from .builder import RuleBuilder
from .common.ref import *

class XmlRules:

    def __init__(self, fpath, gname, rules_node):
        self.gname = gname
        self.fpath = fpath
        self.root = rules_node
        self.__rules = []
        from arjuna.tpi import Arjuna
        self.console = Arjuna.get_console()
        self.__process()


    def __create_rule(self, rule, tobject, rule_type, rcontainer, robject, condition, expression):
        if not robject:
            raise Exception("Target Rule Object not defined.")
        if not condition:
            raise Exception("Target condition not defined.")
        if not expression:
            raise Exception("Target expression not defined.")

        rule_builder = RuleBuilder()
        rule_builder.xml(rule)
        rule_builder.test_object_type(tobject)
        rule_builder.rule_type(rule_type)
        rule_builder.target(rcontainer)
        if (type(robject) is str):
            rule_builder.rule_object(robject.lower())
            rule_builder.condition(condition.upper())
        else:
            rule_builder.rule_object(robject)
            rule_builder.condition(condition)

        rule_builder.expression(expression)
        self.__rules.append(rule_builder.build())

    def __create_4_word_rule_for_container(self, rule, tobject, rule_type, match):
        container = match.group('container').lower()
        if container not in PROPERTY_CONTAINER_MAP:
            raise Exception("Property type in a rule definition can only be one of : {}".format(
                list(PROPERTY_CONTAINER_MAP.keys())
            ))
        self.__create_rule(rule, tobject, rule_type, PROPERTY_CONTAINER_MAP[container], match.group('object'), match.group('condition'),match.group('expression'))

    def __create_3_word_value_rule_for_prop(self, rule, tobject, rule_type, match):
        self.__create_rule(rule, tobject, rule_type, RuleTargetType.PROPS, match.group('object'), match.group('condition'), match.group('expression'))

    def __create_2_word_value_rule_for_tags(self, rule, tobject, rule_type, match):
        tags = [i.strip().lower() for i in match.group('object').split(',')]
        self.__create_rule(rule, tobject, rule_type, RuleTargetType.TAGS, tags, 'is', 'defined')

    def __create_4_word_value_rule_for_multi_defined(self, rule, tobject, rule_type, match):
        container = match.group('container').lower()
        if container not in CONTAINER_MAP:
            raise Exception("Property type in a rule definition for an >>are defined/present<< condition can only be one of : {}".format(
                list(CONTAINER_MAP.keys())
            ))
        tags = [i.strip().lower() for i in match.group('object').split(',')]
        self.__create_rule(rule, tobject, rule_type, CONTAINER_MAP[container], tags,
                           RuleConditionType.CONTAINS, match.group('expression'))

    def __create_1_word_boolean_rule(self, rule, tobject, rule_type, match):
        robject = match.group('object').lower()
        if (mrules.is_builtin_prop(robject) and mrules.get_value_type(robject) is not bool) \
                or \
                (not mrules.is_builtin_prop(robject)):
            raise Exception("One word rules are available only for built-in properties of boolean nature.")
        self.__create_rule(rule, tobject, rule_type, RuleTargetType.PROPS, robject, 'is', 'True')

    def __process(self):
        def display_err_and_exit(msg):
            self.console.display_error((msg + " Fix groups template file: {}").format(self.fpath))
            sys_utils.fexit()

        node_dict = etree_utils.convert_to_cidict(self.root)

        # Validate only pickers keys exist.
        if not set(node_dict.keys()).issubset({'modules', 'functions'}):
            display_err_and_exit(">>rules<< element can contain only >>modules<< and/or >>functions<< elements. Check group: {}.".format(self.gname))
        elif not node_dict:
            display_err_and_exit(">>rules<< element is empty.Check group: {}.".format(self.gname))

        # xyz -- boolean built-in property or flags
        pattern_1 = re.compile(r"^\s*(?P<object>\w+)\s*$",re.IGNORECASE)
        # tagged abc[,.....]
        pattern_2 = re.compile(r"^\s*tagged\s*(?P<object>(.+?))\s*$",re.IGNORECASE)
        # tags/bugs a[,....] are defined
        pattern_3 = re.compile(r"^\s*(?P<container>\w+)\s*(?P<object>.*?)\s+(?P<condition>are)\s+(?P<expression>defined|present)\s*$", re.IGNORECASE)
        # prop is DEFINED
        pattern_4 = re.compile(r"^\s*(?P<object>\w+)\s+(?P<condition>is)\s+(?P<expression>defined|present)\s*$", re.IGNORECASE)
        # xyz condition def
        condition_str = "is|eq|=|==|lt|le|gt|ge|matches|~=|contains|\*="
        p5_raw = r"^\s*(?P<object>\w+)\s+(?P<condition>({}){{1,1}})\s+(?P<expression>.*?)\s*$".format(condition_str)
        pattern_5 = re.compile(p5_raw, re.IGNORECASE)
        # property_type xyz condition def
        condition_str = "is|eq|=|==|lt|le|gt|ge|matches|~=|contains|\*="
        p6_raw = r"^\s*(?P<container>prop|evar|tag|bug)\s*(?P<object>\w+)\s+(?P<condition>({}){{1,1}})\s+(?P<expression>.*?)\s*$".format(condition_str)
        pattern_6 = re.compile(p6_raw, re.IGNORECASE)

        for target_object in list(self.root):
            tobject = target_object.tag.lower() == 'modules' and 'Module' or 'Function'
            for rule in list(target_object):
                if rule.tag not in {'include', 'exclude'}:
                    display_err_and_exit(
                        ">>{}<< element in >>rules<< element can contain only rules with >>include<< and/or >>exclude<< tags. Check group: {}.".format(
                            target_object.tag,
                            self.gname)
                    )
                run_conf_utils.validate_rule_xml_child("groups", self.gname, self.fpath, rule)
                rule_dict = etree_utils.convert_attribs_to_cidict(rule)
                try:
                    match = pattern_1.match(rule_dict['if'])
                    if match:
                        self.__create_1_word_boolean_rule(rule, tobject, rule.tag, match)
                        continue

                    match = pattern_2.match(rule_dict['if'])
                    if match:
                        self.__create_2_word_value_rule_for_tags(rule, tobject, rule.tag, match)
                        continue

                    match = pattern_3.match(rule_dict['if'])
                    if match:
                        self.__create_4_word_value_rule_for_multi_defined(rule, tobject, rule.tag, match)
                        continue

                    match = pattern_4.match(rule_dict['if'])
                    if match:
                        self.__create_3_word_value_rule_for_prop(rule, tobject, rule.tag, match)
                        continue

                    match = pattern_5.match(rule_dict['if'])
                    if match:
                        self.__create_3_word_value_rule_for_prop(rule, tobject, rule.tag, match)
                        continue

                    match = pattern_6.match(rule_dict['if'])
                    if match:
                        self.__create_4_word_rule_for_container(rule, tobject, rule.tag, match)
                        continue

                    # Invalid rule
                    raise Exception("The >>if<< condition content in the rule is invalid: " + rule_dict['if'])

                except Exception as e:
                    import traceback
                    traceback.print_exc()
                    display_err_and_exit(
                        "Exception in processing rule: {}. {}. Check group: {}.".format(ETree.tostring(rule).strip(),e, self.gname))

    def evaluate(self, test_object):
        for rule in self.__rules:
            rule.evaluate(test_object)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import re

def is_true(input):
    if type(input) is not bool:
        raise Exception("Not a boolean")
    return input is True


def is_false(input):
    if type(input) is not bool:
        raise Exception("Not a boolean")
    return input is False

def throw_exception_for_incompt_equal_same(context, actual, expected):
    if actual is None or expected is None:
        return

    if type(actual) != type(expected):
        msg = 'Checking for {} objects does not work for objects of different types.'.format(context)
        raise Exception(msg)

def are_equal(actual, expected):
    throw_exception_for_incompt_equal_same("equal/non-equal", actual, expected)
    return expected == actual

def are_not_equal(actual, expected):
    throw_exception_for_incompt_equal_same("equal/non-equal", actual, expected)
    return not are_equal(actual, expected)

def are_almost_equal(actual, expected, delta):
    diff = abs(expected - actual)
    delta2 = abs(delta)
    return diff <= delta2


def are_same(actual, expected):
    throw_exception_for_incompt_equal_same("same/different", actual, expected)
    return expected is actual


def are_different(actual, expected):
    throw_exception_for_incompt_equal_same("same/different", actual, expected)
    return expected is not actual

def match(target, pattern):
    if target is None: return False
    if re.match('^' + pattern + '$', target):
        return True
    else:
        return False

def partially_match(target, pattern):
    if target is None: return False
    if re.search(pattern, target):
        return True
    else:
        return False

def match_with_ignore_case(target, pattern):
    if target is None: return False
    if re.match('^' + pattern + '$', target, re.IGNORECASE):
        return True
    else:
        return False

def partially_match_with_ignore_case(target, pattern):
    if target is None: return False
    if re.search(pattern, target, re.IGNORECASE):
        return True
    else:
        return False

are_not_same = are_different

def is_none(obj):
    return obj is None


def is_not_none(obj):
    return obj is not None


def contains(parent, child):
    return child in parent

def does_not_contain(parent, child):
    return child not in parent

numbers = {int, float}

def __validate_args_as_numbers(func, left, right):
    if type(left) not in numbers or type(right) not in numbers:
        raise Exception("{} relation check is allowed only for numbers. Received >>{}<< of type: {} and >>{}<< of type {}.".format(
            func,
            left,
            type(left),
            right,
            type(right)
        ))


def less_than(left, right):
    __validate_args_as_numbers("less_than", left, right)
    return left < right

def less_or_equal(left, right):
    __validate_args_as_numbers("less_or_equal", left, right)
    return left <= right

def greater_than(left, right):
    __validate_args_as_numbers("greater_than", left, right)
    return left > right

def greater_or_equal(left, right):
    __validate_args_as_numbers("greater_or_equal", left, right)
    return left >= right



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import sys
import re

from arjuna.tpi.enums import ArjunaOption

def ustr(input):
    return (str(input)).upper()

VNREGEX = r'^[a-z][_a-z0-9]{2,29}$'
VNREGEX_TEXT = '''
{} name must be a string of length 3-30 containing lower case letters, digits or _ (underscore).
It must begin with a letter.
'''

import os


def new_project_dir(input):
    input = input.rstrip("/").rstrip("\\")
    proj_name = os.path.basename(input)
    lname_check("Project", proj_name)
    if os.path.exists(input):
        print('Project path already exist: {}'.format(input))
        return input
    else:
        os.makedirs(input)
        return input

def project_dir(input):
    if not os.path.exists(input):
        print('Project path does not exist: {}'.format(input))
        print('Exiting...', file=sys.stderr)
        sys.exit(1)
    elif not os.path.isdir(input):
        print('Project path is not a directory: {}'.format(input))
        print('Exiting...', file=sys.stderr)
        sys.exit(1)
    else:
        proj_name = os.path.basename(input)
        lname_check("Project", proj_name)
        return input

def lname_check(context, input):
    if not re.match(VNREGEX, input):
        print('Invalid {} name provided.'.format(context), file=sys.stderr)
        print(VNREGEX_TEXT.format(context), file=sys.stderr)
        print('Exiting...', file=sys.stderr)
        sys.exit(1)
    return input


def port(input):
    if type(input) is not str or int(input) < 1024:
        print('Invalid Setu port {} provided. It should be an int such that 1024 < port < 65535'.format(input))
        sys.exit(1)
    return input


from arjuna.setuext.guiauto.dispatcher.automator import GuiAutomatorDispatcher

class TestSessionDispatcher:

    def create_gui_automator_dispatcher(self, config, setu_id):
        return GuiAutomatorDispatcher(config, setu_id)

    def create_gui_element_dispatcher(self, atuomator_dispatcher, element_setu_id):
        return atuomator_dispatcher.create_gui_element_dispatcher(element_setu_id)



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .base import TestObject
from .test import Test
from arjuna.unitee.engine.pool import TestObjectThreadPool
from arjuna.unitee.enums import *
from arjuna.unitee.state.states import *
from arjuna.unitee.reporter.result.types import *
from arjuna.lib.thread import decorators

class TestFunc(TestObject):

	def __init__(self, group, module, mslot, defn):
		super().__init__(TestObjectTypeEnum.Function)
		import threading
		from arjuna.tpi import Arjuna
		self.unitee = Arjuna.get_unitee_instance()
		self.lock = threading.RLock()
		self.group = group
		self.rules = self.group.rules
		self.module = module
		self.mslot = mslot
		self.defn = defn
		self.func_obj = self.defn.func
		self.thcount = self.defn.threads
		self.counter = 0
		self.state = FunctionState()
		self.data_source = None
		self.__dependency = self.defn.dependency
		self.append_before_fixture(self.module.fixture_defs.build(FixtureTypeEnum.INIT_EACH_FUNCTION))
		self.append_after_fixture(self.module.fixture_defs.build(FixtureTypeEnum.END_EACH_FUNCTION))

	def load(self, base_tvars):
		self._populate_tvars(base_tvars)

	@decorators.sync_method('lock')
	def next(self):
		import threading
		try:
			if self.data_source is None:
				self.data_source = self.defn.data_source
		except Exception as e:
			issue = self.unitee.state_mgr.create_issue_from_exception(e, traceback.format_exc(), ResultTypeEnum.ERROR)
			result = TestObjectResult(self, rtype=ResultTypeEnum.ERROR, rcode=ResultCodeEnum.DATASOURCE_CONSTRUCTION_ERROR, iid=issue.iid)
			result.append_issue(issue)
			self.report(result)
			raise SubTestsFinished()

		self.counter += 1
		record = None
		try:
			record = self.data_source.next()
		except DataSourceFinished:
			raise SubTestsFinished()
		else:
			test = Test(self.counter, self.group, self.module, self)
			tvars = self.tvars.clone()
			tvars.data.record = record
			test.load(tvars)
			return test

	def _populate_tvars(self, base_tvars):
		self.tvars = base_tvars
		self.tvars.info.function = self.defn.tvars.info.function
		self.tvars.evars.update(self.defn.tvars.evars)
		self.tvars.tags.update(self.defn.tvars.tags)
		self.tvars.bugs.update(self.defn.tvars.bugs)

	def _execute(self):
		self.tvars.context = self.mslot.tvars.context.clone()
		self.tvars.runtime.update(self.mslot.tvars.runtime)
		tp = TestObjectThreadPool(self.thcount, self, "{}::{}".format(self.thname, "t"))
		tp.run()

	def report(self, result):
		super().report(result)
		self.module.state.update(result)
		self.module.state.update_test_summary_from_state(self.state)

	def _evaluate_dependency(self):
		if self.__dependency:
			self.__dependency.evaluate(self.module.state)

	def _evaluate_rules(self):
		if self.rules is not None:
			self.rules.evaluate(self)


from arjuna.setuext.guiauto.dispatcher.driver.impl.element_commands import DriverElementCommands
from arjuna.setuext.guiauto.dispatcher.driver.impl.element_finder import ElementFinder

class SeleniumDriverElement:

    def __init__(self, driver_wrapper, element_setu_id):
        self.__driver_wrapper = driver_wrapper
        self.__driver = driver_wrapper.driver
        self.__automator_setu_id = driver_wrapper.setu_id
        self.__element_set_id = element_setu_id
        self.__partial = False
        self.__instance_index = 0

    def set_partial(self, index):
        self.__partial = True
        self.__instance_index = index

    @property
    def automator_setu_id(self):
        return self.__automator_setu_id

    @property
    def element_setu_id(self):
        return self.__element_set_id


    def find_element(self, child_gui_element_setu_id, with_type, with_value):
        wrapped_element = self.__driver_wrapper.get_driver_element(self.element_setu_id)
        child_element = ElementFinder.find_element(wrapped_element, with_type, with_value)
        self.__driver_wrapper.add_driver_element(child_gui_element_setu_id, child_element)

    def find_multielement(self, child_gui_element_setu_id, with_type, with_value):
        wrapped_element = self.__driver_wrapper.get_driver_element(self.element_setu_id)
        child_melement = ElementFinder.find_elements(wrapped_element, with_type, with_value)
        self.__driver_wrapper.add_driver_melement(child_gui_element_setu_id, child_melement)
        return child_melement.get_instance_count()

    @property
    def driver(self):
        return self.__driver_wrapper.driver

    @property
    def driver_element(self):
        if not self.__partial:
            return self.__driver_wrapper.get_driver_element(self.element_setu_id)
        else:
            return self.__driver_wrapper.get_driver_melement(self.element_setu_id).get_element_at_index(self.__instance_index)

    def click(self):
        DriverElementCommands.click(self.driver_element)

    def clear_text(self):
        DriverElementCommands.clear_text(self.driver_element)

    def send_text(self, text):
        DriverElementCommands.send_text(self.driver_element, text)

    def is_selected(self):
        return DriverElementCommands.is_selected(self.driver_element)

    def is_visible(self):
        return DriverElementCommands.is_visible(self.driver_element)

    def is_clickable(self):
        return DriverElementCommands.is_clickable(self.driver_element)

    def get_tag_name(self):
        return DriverElementCommands.get_tag_name(self.driver_element)

    def get_attr_value(self, attr_name, optional=False):
        try:
            return DriverElementCommands.get_attr_value(self.driver_element, attr_name)
        except Exception as e:
            if optional:
                return None
            else:
                raise e

    def get_text_content(self):
        return DriverElementCommands.get_text_content(self.driver_element)


from functools import partial

class GuiElementDispatcher:

    def __init__(self, config, automator_name, automator_dispatcher, element_setu_id):
        self.__config = config
        self.__dispatcher = None

        self.__dispatcher = automator_dispatcher.create_gui_element_dispatcher(element_setu_id)

    def __getattr__(self, attr):
        return partial(vars(self.__dispatcher.__class__)[attr], self.__dispatcher)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .base import *
from .gslot import TestGroupSlot
from arjuna.unitee.engine.pool import TestObjectThreadPool
from arjuna.unitee.enums import *
from arjuna.unitee.state.states import *

class TestGroup(TestObject):
	def __init__(self, num, stage, defn):
		super().__init__(TestObjectTypeEnum.Group)
		self.num = num
		self.stage = stage
		self.defn = defn
		self.rules = self.defn.rules
		self.thcount = 1
		self.mthread_count = self.defn.threads
		self.gslots = []
		self.__module_map = None
		self.state = GroupState()
		self.fixture_defs = self.defn.fixture_defs
		self.append_before_fixture(self.stage.fixture_defs.build(FixtureTypeEnum.INIT_EACH_GROUP))
		self.append_before_fixture(self.fixture_defs.build(FixtureTypeEnum.INIT_GROUP))
		self.append_after_fixture(self.fixture_defs.build(FixtureTypeEnum.END_GROUP))
		self.append_after_fixture(self.stage.fixture_defs.build(FixtureTypeEnum.END_EACH_GROUP))

	def load(self, base_tvars):
		self._populate_tvars(base_tvars)
		picked_mnames = self.defn.get_picked_mnames()

		self.__module_map = self.defn.get_schedule_module_map()
		slots = self.unitee.testdb.slot_module_names(picked_mnames)

		for index, mnames in enumerate(slots):
			gslot = TestGroupSlot(index + 1, self, mnames)
			gslot.load(self.tvars.clone())
			self.gslots.append(gslot)

		self.children = self.gslots

	def _populate_tvars(self, base_tvars):
		self.tvars = base_tvars
		self.tvars.info.group.meta['name'] = self.defn.name
		if not self.defn.config.is_empty():
			self.tvars.context.update_with_file_config_container(self.defn.config)

	def _execute(self):
		tp = TestObjectThreadPool(self.thcount, self, "{}::{}".format(self.thname, "gs"))
		tp.run()

	def get_fnames_for_module(self, mname):
		return self.__module_map[mname]

	def report(self, result):
		super().report(result)
		self.stage.state.update(result)
		self.stage.state.update_test_summary_from_state(self.state)


import os
from .source import *

class DataBroker:

    def __init__(self):
        self.__data_sources = {}

    def create_file_data_source(self, data_dir, file_name, record_format="MAP", delimiter="\t"):
        ds = None
        ext = file_name.lower()
        path = os.path.join(data_dir, file_name)
        rformat = record_format.upper()
        if ext.endswith(".csv") or ext.endswith(".txt"):
            if rformat == "LIST":
                ds = DsvFileListDataSource(path, delimiter)
            else:
                ds = DsvFileMapDataSource(path, delimiter)
        elif ext.endswith(".xls"):
            if rformat == "LIST":
                ds = ExcelFileListDataSource(path)
            else:
                ds = ExcelFileMapDataSource(path)
        elif ext.endswith(".ini"):
            ds = IniFileDataSource(path)
        else:
            raise Exception("This is not a default file format supported as a data source: " + path)
        self.__data_sources[ds.setu_id] = ds
        return ds.setu_id

    def get_next_record(self, ds_setu_id):
        ds = self.__data_sources[ds_setu_id]
        return ds.next().record

    def get_all_records(self, ds_setu_id):
        ds = self.__data_sources[ds_setu_id]
        out = []
        while True:
            try:
                record = ds.next().record
                out.append(record)
            except:
                break
        return out

    def reset(self, ds_setu_id):
        ds = self.__data_sources[ds_setu_id]
        return ds.reset()



import sys
import os
from arjuna.main import main

args = sys.argv

if "launch-setu" not in args:
    proj_dir = os.path.dirname(os.path.realpath(__file__))
    args.extend(("-p", proj_dir))
main(args)


import abc
import types
from .guiautomator import GuiAutomator

class Handler(metaclass=abc.ABCMeta):

    def __init__(self, automator: GuiAutomator):
        self.__automator = automator
        self.__config = automator.config

    @property
    def automator(self):
        return self.__automator

    @property
    def config(self):
        return self.__config

    def _act(self, json_dict):
        return self.automator.actor_callable(json_dict)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from .base import *
from .mslot import TestModuleSlot
from arjuna.unitee.engine.pool import TestObjectThreadPool
from arjuna.unitee.enums import *
from arjuna.unitee.state.states import *

class TestModule(TestObject):
	def __init__(self, group, mdef):
		super().__init__(TestObjectTypeEnum.Module)
		self.group = group
		self.rules = self.group.rules
		self.defn = mdef
		self.thcount = 1
		self.fthread_count = self.defn.threads
		self.mslots = []
		self.results = {}
		self.fixture_defs = self.defn.fixture_defs
		self.append_before_fixture(self.group.fixture_defs.build(FixtureTypeEnum.INIT_EACH_MODULE))
		self.append_before_fixture(self.fixture_defs.build(FixtureTypeEnum.INIT_MODULE))
		self.append_after_fixture(self.fixture_defs.build(FixtureTypeEnum.END_MODULE))
		self.append_after_fixture(self.group.fixture_defs.build(FixtureTypeEnum.END_EACH_MODULE))

		self.state = ModuleState()
		self.__dependency = self.defn.dependency

	def load(self, base_tvars):
		self._populate_tvars(base_tvars)
		picked_fnames = self.group.get_fnames_for_module(self.defn.qname)

		from arjuna.unitee import Unitee
		slots = self.defn.slot_func_names(picked_fnames)

		for index, slot in enumerate(slots):
			mslot = TestModuleSlot(index + 1, self.group, self, slot)
			mslot.load(self.tvars.clone())
			self.mslots.append(mslot)
		self.children = self.mslots

	def _populate_tvars(self, base_tvars):
		self.tvars = self.defn.tvars.clone()
		self.tvars.context = base_tvars.context
		self.tvars.runtime.update(base_tvars.evars)
		self.tvars.info.session = base_tvars.info.session
		self.tvars.info.stage = base_tvars.info.stage
		self.tvars.info.group = base_tvars.info.group

	def _execute(self):
		tp = TestObjectThreadPool(self.thcount, self, "{}::{}".format(self.thname, "ms"))
		tp.run()

	def report(self, result):
		super().report(result)
		self.group.state.update(result)
		self.group.state.update_test_summary_from_state(self.state)

	def _evaluate_dependency(self):
		if self.__dependency:
			self.__dependency.evaluate(self.group.state)

	def _evaluate_rules(self):
		if self.rules is not None:
			self.rules.evaluate(self)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.tpi import Arjuna
from arjuna.tpi.enums import ArjunaOption
from arjuna.unitee.ddt.source import *
from arjuna.unitee.ddt.ddt_factory import *
from arjuna.unitee.enums import *
from arjuna.lib.utils import file_utils
from arjuna.lib.utils import obj_utils

class _DataMarkUp:
    pass

class _DataRecord(_DataMarkUp):
    from arjuna.unitee.ddt.source import SingleDataRecordSource
    def __init__(self, *vargs, **kwargs):
        super().__init__()
        if not vargs and not kwargs:
            raise Exception("No data provided in Data")
        self.record = DataRecord(*vargs, **kwargs)

    def build(self):
        source = SingleDataRecordSource(self.record)
        return source

    def get_record(self):
        return self.record

record = _DataRecord

class _DataRecords(_DataMarkUp):
    from arjuna.unitee.ddt.source import DataArrayDataSource
    def __init__(self, *records):
        super().__init__()
        for i in records:
            if not isinstance(i, record):
                raise Exception("Items in records() must be instances of record.")
        self.records = [i.get_record() for i in records]

    def build(self):
        source = DataArrayDataSource(self.records)
        return source

records = _DataRecords

class _DataFile(_DataMarkUp):
    def __init__(self, path=None, delimiter="\t"):
        super().__init__()
        self.path = path
        self.delimiter = delimiter

        data_dir = Arjuna.get_central_config().get_arjuna_option_value(ArjunaOption.DATA_SOURCES_DIR).as_string()

        if file_utils.is_absolute_path(self.path):
            if not file_utils.is_file(self.path):
                if file_utils.is_dir(self.path):
                    raise Exception("Not a file: {}".format(self.path))
                else:
                    raise Exception("File does not exist: {}".format(self.path))
        else:
            self.path = os.path.abspath(os.path.join(data_dir, self.path))
            if not file_utils.is_file(self.path):
                if file_utils.is_dir(self.path):
                    raise Exception("Not a file: {}".format(self.path))
                else:
                    raise Exception("File does not exist: {}".format(self.path))

    def build(self):
        source = create_ds(self.path, self.delimiter)
        return source

data_file = _DataFile

class _DataFunc(_DataMarkUp):
    def __init__(self, func, *vargs, **kwargs):
        super().__init__()
        if not obj_utils.callable(func):
            raise Exception("Argument provided to data_function should be a function object.")
        elif not obj_utils.is_function(func):
            raise Exception("Argument for data_function should be a function object which does not exist inside a class.")
        self.func = func
        self.vargs = vargs
        self.kwargs = kwargs

    def build(self):
        return DataFunctionDataSource(self.func, *self.vargs, **self.kwargs)

data_function = _DataFunc

class _DataClass(_DataMarkUp):
    def __init__(self, dsclass, *vargs, **kwargs):
        super().__init__()
        if not isinstance(type(dsclass), type):
            raise Exception("Data Class expects dsclass to be a class.")
        self.dsclass = dsclass
        self.vargs = vargs
        self.kwargs = kwargs

    def build(self):
        return DataClassDataSource(self.dsclass, *self.vargs, **self.kwargs)

data_class = _DataClass

class _MultiDataSource(_DataMarkUp):
    def __init__(self, *dsources):
        super().__init__()
        if not dsources:
            raise Exception("You must provide atleast one data source as argument for many_data_sources")
        for dsource in dsources:
            if not isinstance(dsource, _DataMarkUp):
                raise Exception("All arguments of many_data_sources must be data sources.")
        self.dsource_defs = dsources

    def build(self):
        return MultiDataSource(self.dsource_defs)

many_data_sources = _MultiDataSource


import base64
import os
import time

from arjuna.tpi.enums import ArjunaOption
from arjuna.setuext.guiauto.impl.base.element_container import ElementContainer
from .drivercaps import DriverCapabilities


class GuiAutomator(ElementContainer):

    def __init__(self, config, extended_config=None):
        super().__init__(config)
        self.__extended_config = extended_config
        self.__automator_uri = "/s02guiauto/automator/{}".format(self.get_setu_id())
        self.__create_screenshots_dir()
        self.__main_window = None
        self.__in_slomo = config.setu_config.value(ArjunaOption.GUIAUTO_SLOMO_ON)
        self.__slomo_interval = config.setu_config.value(ArjunaOption.GUIAUTO_SLOMO_INTERVAL)

        from .webalert_handler import WebAlertHandler
        from .automator_conditions import GuiAutomatorConditions
        from .viewcontext_handler import ViewContextHandler
        self.__alert_handler = WebAlertHandler(self)
        self.__conditions_handler = GuiAutomatorConditions(self)
        self.__view_handler = ViewContextHandler(self)
        self.__browser = None

    def _create_element_flat_or_nested(self, locator_meta_data):
        from arjuna.setuext.guiauto.impl.element.guielement import GuiElement
        return GuiElement(self, locator_meta_data) 

    def _create_multielement_flat_or_nested(self, locator_meta_data):
        from arjuna.setuext.guiauto.impl.element.multielement import GuiMultiElement
        return GuiMultiElement(self, locator_meta_data) 

    def create_dispatcher(self):
        self._set_dispatcher(self.dispatcher_creator.create_gui_automator_dispatcher(self.config, self.setu_id))

    def slomo(self):
        if self.__in_slomo:
            time.sleep(self.__slomo_interval)

    def set_slomo(self, on, interval=None):
        self.__in_slomo = on
        if interval is not None:
            self.__slomo_interval = interval

    @property
    def browser(self):
        return self.__browser

    @property
    def main_window(self):
        return self.__main_window

    @property
    def alert_handler(self):
        return self.__alert_handler

    @property
    def view_handler(self):
        return self.__view_handler

    @property
    def conditions(self):
        return self.__conditions_handler

    def __create_screenshots_dir(self):
        sdir = self.config.setu_config.value(ArjunaOption.SCREENSHOTS_DIR)
        if not os.path.isdir(sdir):
            os.makedirs(sdir)

    #Override
    def _get_object_uri(self):
        return self.__automator_uri

    def launch(self):
        caps = DriverCapabilities(self.config, self.__extended_config)
        self.dispatcher.launch(caps.processed_config)

        from arjuna.setuext.guiauto.impl.element.window import MainWindow
        self.__main_window = MainWindow(self)

        from .browser import Browser
        self.__browser = Browser(self)

    def quit(self):
        self.dispatcher.quit()

    def __screenshot(self):
        switch_view_context = None
        if self.config.value(ArjunaOption.MOBILE_OS_NAME).lower() == "android":
            view_name = self.view_handler.get_current_view_context()   
            if self.view_handler._does_name_represent_web_view(view_name) :
                self.view_handler.switch_to_native_view() 
                switch_view_context = view_name

        response = self.dispatcher.take_screenshot()

        if switch_view_context:
            self.view_handler.switch_to_view_context(switch_view_context)
        
        return response

    def take_screenshot(self):
        response = self.__screenshot()
        image = base64.b64decode(response["data"]["codedImage"])
        path = os.path.join(self.config.value(ArjunaOption.SCREENSHOTS_DIR), "{}.png".format(str(time.time()).replace(".", "-")))
        f = open(path, "wb")
        f.write(image)
        f.close()

    def focus_on_main_window(self):
        self.main_window.focus()

    def add_frame(self, frame):
        self._add_element(frame.setu_id, frame)



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import types
import copy
import pprint

from arjuna.lib.adv.types import *
from arjuna.lib.adv.proxy import ROProxy
from arjuna.lib.types.descriptors import *
from arjuna.unitee.validation.steps import Steps
from arjuna.lib.config import InternalTestContext

SCALARS = {int, float, str, bool}

class SingleObjectVars(CIStringDict):

    def __setitem__(self, key, value):
        if type(value) not in SCALARS:
            raise Exception("Single Vars dictionaries can only contain scalars ({}) as values.".format(SCALARS))
        super().__setitem__(key, value)

    def merge_with_parent(self, parent_vars):
        _d = SingleObjectVars()
        _d.update(parent_vars)
        _d.update(self)
        self.update(_d)

class evars:
    def __init__(self, **kwargs):
        self.__dict = kwargs
        
    def get_dict(self):
        return self.__dict

    def __str__(self):
        return pprint.pformat(self.__dict)


class MultiVars:

    def __init__(self, child_count, **kwargs):
        self.__svars = [SingleObjectVars() for i in range(child_count)]
        for k,v in kwargs.items():
            bl = self.__balance(child_count, v)
            for i in range(child_count):
                self.__svars[i][k] = bl[i]

    def __balance(self, count, l):
        if not ListTupleOrSet.check(l):
            l = [l]
        for i in l:
            if type(i) not in SCALARS:
                raise Exception("Multi Vars dictionaries can only contain list of scalars ({}) as values.".format(SCALARS))
        if len(l) == count:
            return l
        elif len(l) > count:
            return l[:count]
        else:
            return l + [l[-1] for i in range(count - len(l))]

    def __str__(self):
        return pprint.pformat(self.__svars)


class TagSet:

    def __init__(self, *vargs):
        self.__set = {str(i).lower() for i in vargs}

    def __fv(self, item):
        return str(item).lower()

    def __contains__(self, item):
        return self.__fv(item) in self.__set

    def update(self, items):
        for item in items:
            self.__set.add(self.__fv(item))

    def __iter__(self):
        return iter(self.__set)

    def is_empty(self):
        return len(self.__set) == 0

    def __str__(self):
        return str(self.__set)

tags = TagSet
bugs = TagSet


class DataRef:
    pass


class DataRefs:
    def __init__(self, **kwargs):
        self._drefs = SingleObjectVars()
        for k,v in kwargs:
            if not isinstance(v, DataRef):
                raise Exception("The argumens of DataRefs must be name=DataRef object.")
            else:
                self.add(k, v)

    def __getattr__(self, item):
        return self._drefs[item]

    def __setattr__(self, key, value):
        if key != "_drefs":
            vars(self)['_drefs'][key] = value
        else:
            vars(self)['_drefs'] = value

    def add(self, name, dref):
        self._drefs[name] = dref

    def __str__(self):
        return pprint.pformat(self._drefs, indent=4)


class TestInfo:
    def __init__(self, meta):
        self.meta = meta
        self.props = CIStringDict()

    def set_props(self, sdict):
        for k,v in sdict.items():
            if k != "kwargs":
                self.props[k] = v
            else:
                for i,j in v.items():
                    self.props[i] = j

    def freeze(self):
        self.meta = types.MappingProxyType(self.meta)
        self.props = types.MappingProxyType(self.props)

    def __str__(self):
        return '''Meta\n{}\nProps\n{}\n'''.format(str(self.meta), str(self.props))

    def clone(self):
        info = self.__class__()
        info.meta = self.meta.clone()
        info.props = ROProxy(self.props.clone())
        return info


class SessionInfo(TestInfo):
    def __init__(self):
        super().__init__(CIStringDict({
            "name": "-"
        }))


class StageInfo(TestInfo):
    def __init__(self):
        super().__init__(CIStringDict({
            "name": "-"
        }))


class GroupInfo(TestInfo):
    def __init__(self):
        super().__init__(CIStringDict({
            "name": "-",
            "slot": "-"
        }))


class ModuleInfo(TestInfo):
    def __init__(self):
        super().__init__(CIStringDict({
            "pkg": "-",
            "name": "-",
            "qname": "-",
            "slot" : "-"
        }))


class FunctionInfo(TestInfo):
    def __init__(self):
        super().__init__(CIStringDict({
            "name": "-",
            "qname": "-"
        }))

# class RODataRecord:
#
#     def __init__(self, drec):
#         if drec:
#             self.indexed = tuple(drec.indexed)
#             self.named = types.MappingProxyType(drec.named)
#         else:
#             self.indexed = None
#             self.named = None


class AllInfo:

    def __init__(self):
        self.session = None
        self.stage = None
        self.group = None
        self.module = None
        self.function = None
        self.object_type = None
        self.test_num = None

    def create_ro_wrapper(self):
        info = AllInfo()
        info.session = ROProxy(self.session)
        info.stage = ROProxy(self.stage)
        info.group = ROProxy(self.group)
        info.module = ROProxy(self.module)
        info.function = ROProxy(self.function)
        info.object_type = self.object_type
        info.test_num = self.test_num
        return ROProxy(info)

    def clone(self):
        info = AllInfo()
        info.session = self.session and self.session.clone() or SessionInfo()
        info.stage = self.stage and self.stage.clone() or StageInfo()
        info.group = self.group and self.group.clone() or GroupInfo()
        info.module = self.module and self.module.clone() or ModuleInfo()
        info.function = self.function and self.function.clone() or FunctionInfo()
        info.object_type = self.object_type
        info.test_num = self.test_num
        return info


class State:

    steps = Steps


class Data:

    def __init__(self):
        self.record = None


class TestVars:

    def __init__(self):
        self.info = AllInfo()
        self.context = None
        self.evars = SingleObjectVars()
        self.runtime = OnceOnlyKeyCIStringDict()
        self.tags = tags() # return frozen set
        self.bugs = tags()  # return frozen set
        self.data = None

    def clone(self):
        nvars = TestVars()
        if self.context is not None:
            nvars.context = self.context.clone()
        else:
            nvars.context = None
        nvars.info = self.info.clone()
        nvars.evars = SingleObjectVars({i:j for i,j in self.evars.items()})
        nvars.tags = TagSet(*{i for i in self.tags})
        nvars.bugs = TagSet(*{i for i in self.bugs})
        nvars.data = Data() #RODataRecord For .pos it should be a tuple, for .map immutable dict
        nvars.runtime = self.runtime.clone()
        return nvars

    def create_utvars(self):
        class UTVars:
            def __init__(self, tvars):
                # RO
                self.info = tvars.info.create_ro_wrapper()
                self.tags = frozenset(tvars.tags)
                self.bugs = frozenset(tvars.bugs)
                self.context = tvars.context.clone_for_user()
                # R/W
                self.evars = tvars.evars
                self.runtime = tvars.runtime
                # data
                self.data = ROProxy(tvars.data)
                # steps
                self.steps = Steps

        return ROProxy(UTVars(self))

    def __str__(self):
        out = ""
        names = list(vars(self).keys())
        names.sort()
        for name in names:
            out += "Contents of: " + name + "\n"
            out += str(vars(self)[name]) + "\n"
        return out




'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from enum import Enum, auto

class DataSourceTypeEnum(Enum):
    DATA = auto()
    DATA_ARRAY = auto()
    DATA_METHOD = auto()
    DATA_FILE = auto()
    DATA_GENERATOR = auto()

class UniteeComponentEnum(Enum):
    TEST_RUNNER = auto()
    TEST_FILTER = auto()

class IssueSubType(Enum):
    pass

class ResultCodeEnum(Enum):
    ERROR_INIT_MODULE = auto()
    ERROR_END_MODULE = auto()
    ERROR_INIT_EACH_FUNCTION = auto()
    ERROR_END_EACH_FUNCTION = auto()
    ERROR_INIT_EACH_TEST = auto()
    ERROR_END_EACH_TEST = auto()
    FAIL_INIT_MODULE = auto()
    FAIL_END_MODULE = auto()
    FAIL_INIT_EACH_FUNCTION = auto()
    FAIL_END_EACH_FUNCTION = auto()
    FAIL_INIT_EACH_TEST = auto()
    FAIL_END_EACH_TEST = auto()

    FAIL_CHILD_TEST_OBJECT = auto()
    ERROR_CHILD_TEST_OBJECT = auto()
    EXCLUDED_CHILD_TEST_OBJECT = auto()
    PASS_ALL_CHILD_TEST_OBJECTS = auto()

    PASS_ALL_STEPS = auto()

    DATASOURCE_CONSTRUCTION_ERROR = auto()
    DEPENDENCY_NOT_MET = auto()
    PARENT_DEPENDENCY_NOT_MET = auto()

    SURRENDERED_RULES_NOT_MET = auto()
    PARENT_SURRENDERED = auto()

    STEP_FAILURE = auto()
    STEP_ERROR = auto()
    SKIPPED_RECORD = auto()
    TEST_CONTAINER_DEPENDENCY_NOTMET = auto()
    tfunc_DEPENDENCY_NOTMET = auto()
    TEST_CONTAINER_CONSTRUCTOR_ERROR = auto()
    ERROR_IN_BEFORE_CLASS = auto()
    ERROR_IN_BEFORE_CLASS_INSTANCE = auto()
    ERROR_IN_BEFORE_CLASS_FRAGMENT = auto()
    ERROR_IN_BEFORE_METHOD = auto()
    ERROR_IN_BEFORE_METHOD_INSTANCE = auto()
    ERROR_IN_BEFORE_TEST = auto()
    ERROR_IN_AFTER_CLASS = auto()
    ERROR_IN_AFTER_CLASS_INSTANCE = auto()
    ERROR_IN_AFTER_CLASS_FRAGMENT = auto()
    ERROR_IN_AFTER_METHOD = auto()
    ERROR_IN_AFTER_METHOD_INSTANCE = auto()
    ERROR_IN_AFTER_TEST = auto()

    dsource_NEXT_ERROR = auto()

class IgnoredTestStatusEnum(Enum):
    UNPICKED = auto()
    SKIPPED = auto()

class IgnoredTestReasonEnum(Enum):
    UNPICKED_CLASS = auto()
    UNPICKED_METHOD = auto()
    SKIPPED_CLASS_ANNOTATION = auto()
    SKIPPED_METHOD_ANNOTATION = auto()

class DeferredReporterNames(Enum):
    EXCEL = auto()
    JXML = auto()

class UnpickedCodeEnum(Enum):
    UNPICKED_CLASS = auto()
    UNPICKED_METHOD = auto()

class ActiveReporterNames(Enum):
    MIN_CONSOLE = auto()
    CONSOLE = auto()
    EXCEL = auto()


class DataRecordOrderEnum(Enum):
    ORDERED = auto()
    UNORDERED = auto()


class SkipCodeEnum(Enum):
    SKIP_MODULE_DEC = auto()
    SKIP_FUNC_DEC = auto()
    SKIP_REC_DEC = auto()

class UserDefinedPropertyEnum(Enum):
    UNKNOWN = auto()

class UniteePropertyEnum(Enum):
    SESSION_NAME = auto()
    RUNID = auto()
    IRUNID = auto()
    FAILFAST = auto()
    PROJECT_NAME = auto()
    PROJECT_DIR = auto()
    PROJECT_CONFIG_DIR = auto()
    TEST_MODULE_IMPORT_PREFIX = auto()
    CONF_FIXTURES_IMPORT_PREFIX = auto()
    DATA_DIR = auto()
    DATA_SOURCES_DIR = auto()
    DATA_REFERENCES_DIR = auto()
    SCREENSHOTS_DIR = auto()
    SESSIONS_DIR = auto()
    GROUPS_DIR = auto()
    TESTS_DIR = auto()
    REPORT_DIR = auto()
    ARCHIVES_DIR = auto()
    CORE_DIR = auto()
    CORE_DB_CENTRAL_DIR = auto()
    CORE_DB_CENTRAL_DBFILE = auto()
    CORE_DB_RUN_DIR = auto()
    CORE_DB_RUN_DBFILE = auto()
    CORE_DB_TEMPLATE_DIR = auto()
    CORE_DB_TEMPLATE_CENTRAL_DBFILE = auto()
    CORE_DB_TEMPLATE_RUN_DBFILE = auto()
    RUN_REPORT_DIR = auto()
    RUN_REPORT_JDB_DIR = auto()
    RUN_REPORT_JSON_DIR = auto()
    RUN_REPORT_JSON_TESTS_DIR = auto()
    RUN_REPORT_JSON_ISSUES_DIR = auto()
    RUN_REPORT_JSON_IGNOREDTESTS_DIR = auto()
    RUN_REPORT_JSON_EVENTS_DIR = auto()
    RUN_REPORT_JSON_FIXTURES_DIR = auto()
    REPORT_NAME_FORMAT = auto()
    ACTIVE_REPORTERS = auto()
    DEFERRED_REPORTERS = auto()
    OFFLINE_REPORTERS = auto()

class DependencyConditionEnum(Enum):
    NO_ISSUES = auto()
    NONE = auto()

class IssueTypeEnum(Enum):
    FIXTURE_STEP_ISSUE = auto()
    DATA_SOURCE_ISSUE = auto()
    TEST_STEP_ISSUE = auto()

class FixtureTypeEnum(Enum):
    INIT_SESSION = auto()
    END_SESSION = auto()
    INIT_EACH_STAGE = auto()
    END_EACH_STAGE = auto()
    INIT_STAGE = auto()
    END_STAGE = auto()
    INIT_EACH_GROUP = auto()
    END_EACH_GROUP = auto()
    INIT_GROUP = auto()
    END_GROUP = auto()
    INIT_EACH_MODULE = auto()
    END_EACH_MODULE = auto()
    INIT_MODULE = auto()
    INIT_EACH_FUNCTION = auto()
    INIT_EACH_TEST = auto()
    END_EACH_TEST = auto()
    END_EACH_FUNCTION = auto()
    END_MODULE = auto()

class IssueSubTypeEnum(Enum):
    FIXTURE_BEFORE_SESSION = auto()
    FIXTURE_AFTER_SESSION = auto()
    FIXTURE_BEFORE_GROUP = auto()
    FIXTURE_AFTER_GROUP = auto()
    FIXTURE_BEFORE_CLASS = auto()
    FIXTURE_AFTER_CLASS = auto()
    FIXTURE_BEFORE_CLASS_INSTANCE = auto()
    FIXTURE_AFTER_CLASS_INSTANCE = auto()
    FIXTURE_BEFORE_CLASS_FRAGMENT = auto()
    FIXTURE_AFTER_CLASS_FRAGMENT = auto()
    FIXTURE_BEFORE_METHOD = auto()
    FIXTURE_AFTER_METHOD = auto()
    FIXTURE_BEFORE_METHOD_INSTANCE = auto()
    FIXTURE_AFTER_METHOD_INSTANCE = auto()
    FIXTURE_BEFORE_TEST = auto()
    FIXTURE_AFTER_TEST = auto()
    CONSTRUCTOR = auto()
    STEP_FAILURE = auto()
    STEP_ERROR = auto()
    dsource_CONSTRUCTION = auto()
    dsource_NEXT_EXCEPTION = auto()

class PickerRuleEnum(Enum):
    CM = auto()
    IM = auto()
    CF = auto()
    IF = auto()

class ResultTypeEnum(Enum):
    PASS = auto()
    FAIL = auto()
    ERROR = auto()
    EXCLUDED = auto()
    EXCLUDED_CHILDREN = auto()
    UNPICKED = auto()
    SKIPPED = auto()
    NOT_A_TEST = auto()
    NOTHING = auto()
    SURRENDERED = auto()

class ReportableType(Enum):
    INFO = auto()
    RESULT = auto()
    ISSUE = auto()

class TestObjectTypeEnum(Enum):
    Session = auto()
    Stage = auto()
    Group = auto()
    GSlot = auto()
    Module = auto()
    MSlot = auto()
    Function = auto()
    Test = auto()
    Fixture = auto()
    Step = auto()

class InfoType(Enum):
    STARTED = auto()
    FINISHED = auto()

class InfoEntryEnum(Enum):
    INFO_TIMESTAMP = auto()
    OTYPE = auto()
    MESSAGE = auto()
    SESSION = auto()
    STAGE = auto()
    GROUP = auto()
    PKG = auto()
    MODULE = auto()
    FUNCTION = auto()
    TEST = auto()
    FIXTURE = auto()
    FIXTYPE = auto()
    BTSTAMP = auto()
    ETSTAMP = auto()
    EXEC_TIME = auto()

class ResultEntryEnum(Enum):
    RESULT_TIMESTAMP = auto()
    OTYPE = auto()
    SESSION = auto()
    STAGE = auto()
    GROUP = auto()
    PKG = auto()
    MODULE = auto()
    FUNCTION = auto()
    TEST = auto()
    FIXTURE = auto()
    FIXTYPE = auto()
    RTYPE = auto()
    RCODE = auto()
    PROPS = auto()
    RUNTIME = auto()
    EVARS = auto()
    TAGS = auto()
    DATA_RECORD = auto()
    IID = auto()
    THREAD_ID = auto()

class ExecutionContext(Enum):
    Fixture = auto()
    Test = auto()

class StepType(Enum):
    ExecStep = auto()
    AssertStep = auto()
    LogStep = auto()

class StepEntryEnum(Enum):
    OTYPE = auto()
    STEP_TIMESTAMP = auto()
    STEP_ID = auto()
    STEP_EXEC_CONTEXT = auto()
    STEP_TYPE = auto()
    STEP_SOURCE = auto()
    PURPOSE = auto()
    STEP_RTYPE = auto()
    IID = auto()
    OBSERVATION = auto()
    EXPECTATION = auto()
    ASSERT_MESSAGE = auto()
    STEP_PROPS = auto()

class IssueEntryEnum(Enum):
    ISSUE_TIMESTAMP = auto()
    IID = auto()
    ITYPE = auto()
    ISTYPE = auto()
    ENAME = auto()
    EMESSAGE = auto()
    ETRACE = auto()
    RESULT_TIMESTAMP = auto()
    OTYPE = auto()
    MESSAGE = auto()
    SESSION = auto()
    STAGE = auto()
    GROUP = auto()
    PKG = auto()
    MODULE = auto()
    FUNCTION = auto()
    TEST = auto()
    FIXTURE = auto()
    FIXTYPE = auto()
    RTYPE = auto()
    RCODE = auto()
    PROPS = auto()
    RUNTIME = auto()
    EVARS = auto()
    TAGS = auto()
    DATA_RECORD = auto()
    STEP_TIMESTAMP = auto()
    STEP_ID = auto()
    STEP_EXEC_CONTEXT = auto()
    STEP_TYPE = auto()
    STEP_SOURCE = auto()
    PURPOSE = auto()
    STEP_RTYPE = auto()
    OBSERVATION = auto()
    EXPECTATION = auto()
    ASSERT_MESSAGE = auto()
    STEP_PROPS = auto()
    THREAD_ID = auto()

class BuiltInProp(Enum):
    ID = auto()
    PRIORITY = auto()
    THREADS= auto()
    NAME = auto()
    AUTHOR = auto()
    IDEA = auto()
    UNSTABLE = auto()
    COMPONENT = auto()
    APP_VERSION = auto()


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import importlib
from functools import partial
import inspect
import traceback

from arjuna.unitee.test.defs.module import *
from arjuna.unitee.test.defs.func import *
from arjuna.unitee.loader import kfactory
from arjuna.unitee.types.containers import *
from arjuna.unitee.enums import *
from arjuna.unitee.markup import mrules

class ModuleLoader:
    def __init__(self, pkg, module, qname):
        self.pkg = pkg
        self.module = module
        self.qname = qname
        from arjuna.tpi import Arjuna
        self.console = Arjuna.get_console()
        self.logger = Arjuna.get_logger()
        self.mdef = ModDef(pkg, module, qname)
        self._current_kall_ids = {}
        self._current_kall_decs = {}
        self.__multidecs = {}
        from arjuna.tpi import Arjuna
        self.unitee = Arjuna.get_unitee_instance()

    def load(self):
        try:
            self.logger.debug("Importing Module in tests directory: {}".format(self.qname))
            importlib.import_module(self.qname)
        except Exception as e:
            self.console.display_error(
                "Not able to import module: {}. Fix the errors leading to following exception.".format(self.qname))
            self.console.display_exception_block(e, traceback.format_exc())
            sys_utils.fexit()

        self.unitee.testdb.register_mdef(self.mdef)

    def __str(self, obj):
        try:
            return ".".join([obj.__module__, obj.__qualname__])
        except:
            return str(obj)

    def __add_to_multi_dec(self, kallable, dec_name):
        qname = kallable.__qualname__
        if qname not in self.__multidecs:
            self.__multidecs[qname] = []
        self.__multidecs[qname].append(dec_name)

    def __isfunc(self, dtype, dec_name, obj):
        if not obj_utils.is_function(obj):
            self.console.display_error("{} can only be an unbound functions. You have decorated {} with @{}".format(
                dtype,
                self.__str(obj),
                dec_name
            )
            )
            sys_utils.fexit()

    def __isinit(self, dtype, dec_name, obj):
        if obj.__name__ != "__init__":
            self.console.display_error(
                "@tmodule can be used only on a function with name __init__. You have decorated {} with @{}".format(
                    self.__str(obj),
                    dec_name
                )
            )
            sys_utils.fexit()

    def __isnotinit(self, dtype, dec_name, obj):
        if obj.__name__ == "__init__":
            self.console.display_error(
                "__init__ is a reserved function name to initialize the module with @tmodule. You have decorated {} with @{}".format(
                    self.__str(obj),
                    dec_name
                )
            )
            sys_utils.fexit()

    def __ispublic(self, dtype, dec_name, obj):
        if not obj_utils.is_public(obj):
            self.console.display_error(
                "{} should be named as per public convention. Its name can not start with _ and __. You have decorated {} with @{}".format(
                    dtype,
                    self.__str(obj),
                    dec_name
                )
            )
            sys_utils.fexit()

    def __ismysig(self, dtype, dec_name, obj):
        if not obj_utils.is_my_sig(obj):
            msg = "The signature of {} should be def funcname(my). You have decorated {} with @{}"
            msg += "The signature of your function is {}. Correct and proceed."
            self.console.display_error(msg.format(dtype, self.__str(obj), dec_name, inspect.getfullargspec(obj)))
            sys_utils.fexit()

    def __wrongmultientry(self, mqname, fname, dstr):
        msg = "Fatal Error: Unsupported multi-decoration in module {}."
        msg += "{} has been decorated with {}."
        self.console.display_error(msg.format(mqname, fname, dstr))
        sys_utils.fexit()

    def __duplicatefunc(self, mqname, fname, dstr):
        msg = "Fatal Error: Duplicate decorate function names found in module: {}."
        msg += "So far, two separate definitions of function {} were found decorated with {}."
        self.console.display_error(msg.format(mqname, fname, dstr))
        sys_utils.fexit()

    def __validate(self, vfuncs, dtype, dec_name, obj):
        for vfunc in vfuncs:
            vfunc(dtype, dec_name, obj)

    def __register_dec(self, dtype, dec_name, kallable, *, initcheck, fnoentry, fmultidec):
        initcheck(dtype, dec_name, kallable)
        kqname = kallable.__qualname__
        id_dict = self._current_kall_ids
        if kqname in id_dict:
            dstr = ", ".join(["@{}".format(d) for d in ([dec_name] + self._current_kall_decs[kqname])])
            if id(kallable) == id_dict[kqname]:
                self.logger.debug("Multi-decorator situation: {} {} {} {} {}".format(dtype, dec_name, kallable, id(kallable), id_dict[kqname]))
                fmultidec(kallable.__module__, kallable.__qualname__, dstr)
            else:
                self.logger.fatal("Duplicate name problem: {} {} {} {} {}".format(dtype, dec_name, kallable, id(kallable), id_dict[kqname]))
                self.__duplicatefunc(kallable.__module__, kallable.__qualname__, dstr)
        else:
            self.logger.debug("First decorator: {} {} {} {}".format(dtype, dec_name, kallable, id(kallable)))
            fnoentry()
        return kallable

    def __update_tracking(self, kallable, dec_name):
        kqname = kallable.__qualname__
        id_dict = self._current_kall_ids
        dec_dict = self._current_kall_decs
        id_dict[kqname] = id(kallable)
        if kqname not in dec_dict:
            dec_dict[kqname] = []
        dec_dict[kqname].append(dec_name)

    def register_tmodule(self, kallable, **tsargs):
        self.register_fixture("init_module", kallable)

        self.mdef.threads = tsargs['threads']
        del tsargs['threads']
        self.mdef.set_evars(tsargs['evars'])
        del tsargs['evars']
        if tsargs['tags']:
            self.mdef.tvars.tags = tsargs['tags']
        del tsargs['tags']
        if tsargs['bugs']:
            self.mdef.tvars.bugs = tsargs['bugs']
        del tsargs['bugs']
        self.mdef.drive_with = tsargs['drive_with']
        del tsargs['drive_with']
        self.mdef.dependency = tsargs['exclude_if']
        del tsargs['exclude_if']

        # Validate props
        try:
            mrules.validate_built_in_props(tsargs)
        except Exception as e:
            raise Exception("Error in @init_module decorator on function: {}.{}. {}.".format(self.qname,kallable, e))
        self.mdef.tvars.info.module.set_props(tsargs)

    def register_tfunc(self, kallable, **tsargs):
        def register_tfunc():
            self.__update_tracking(kallable, "test_function")
            t = kfactory.create_test_func(kallable)
            td = FuncDef(self.mdef, t)

            td.threads = tsargs['threads']
            del tsargs['threads']
            td.set_evars(tsargs['evars'])
            del tsargs['evars']
            if tsargs['tags']:
                td.tvars.tags = tsargs['tags']
            del tsargs['tags']
            if tsargs['bugs']:
                td.tvars.bugs = tsargs['bugs']
            del tsargs['bugs']
            td.data_driver = tsargs['drive_with']
            del tsargs['drive_with']
            td.data_ref = tsargs['data_ref']
            del tsargs['data_ref']
            td.dependency = tsargs['exclude_if']
            del tsargs['exclude_if']

            # Validate props
            try:
                mrules.validate_built_in_props(tsargs)
            except Exception as e:
                raise Exception("Error in @test decorator of test module: {}. {}.".format(t.qname, e))
            mrules.validate_built_in_props(tsargs)
            td.tvars.info.function.set_props(tsargs)

            self.mdef.register_fdef(td)

        return self.__register_dec(
            "Test function",
            "tfunc",
            kallable,
            initcheck = partial(self.__validate, [self.__isfunc, self.__isnotinit, self.__ispublic, self.__ismysig]),
            fnoentry = register_tfunc,
            fmultidec = self.__wrongmultientry
        )

    def register_fixture(self, dec_name, kallable):
        def register_fixture():
            self.__update_tracking(kallable, dec_name)
            f = kfactory.create_fixture(dec_name, kallable)
            self.mdef.add_fixture_func(f.type, f)

        return self.__register_dec(
            "Fixture function",
            dec_name,
            kallable,
            initcheck= partial(self.__validate, [self.__isfunc, self.__isnotinit, self.__ispublic, self.__ismysig]),
            fnoentry = register_fixture,
            fmultidec = self.__wrongmultientry
        )

    def register_skip(self, kallable):
        def not_a_test():
            msg = "Fatal Error: You have used @skip instruction on a non-test function. "
            msg += "Either @tfunc or @tmodule decorators are missing or have been put in wrong order. "
            msg += "Use @skip\n@tmodule or @skip\@tfunc as applicable."
            self.console.display_error(msg)
            sys_utils.fexit()

        def register_skip(*vargs):
            self.__update_tracking(kallable, "skip")
            if self.mdef.get_fixture_name(FixtureTypeEnum.INIT_MODULE) == kallable.__name__:
                self.mdef.set_skip_code(SkipCodeEnum.SKIP_MODULE_DEC)
            else:
                td = self.mdef.get_fdef(kallable.__qualname__)
                td.set_skip_code(SkipCodeEnum.SKIP_FUNC_DEC)

        return self.__register_dec(
            "Skipped function",
            "skip",
            kallable,
            initcheck = partial(self.__validate, [self.__isfunc, self.__ispublic, self.__ismysig]),
            fnoentry = not_a_test,
            fmultidec = register_skip
        )

    def enumerate(self):
        self.console.display("M" * 50)
        self.console.display("Module: ", self.qname)
        self.mdef.enumerate()
        self.console.display("M" * 50)


import logging
import sys
import inspect
import time

from arjuna.setuext.guiauto.impl.automator.guiautomator import GuiAutomator
from arjuna.setuext.guiauto.impl.locator.emd import GuiElementMetaData, SimpleGuiElementMetaData
from functools import partial

class _DummyLogger:

    def dummy(self, item, *vargs):
        print("Dummy call", item)
        frame = inspect.stack()[3]
        caller_str = "{}.py:{}:L{}".format(inspect.getmodule(frame[0]).__name__, frame[3], frame[2])
        print("Dummy logger called by {} with args: ".format(caller_str), str(vargs)[:300], "...")
        pass

    def __getattr__(self, item):
        return partial(self.dummy, item)


class Setu:
    LOGGER = None
    __DUMMY_LOGGER = _DummyLogger()

    @classmethod
    def init_logger(cls, testsession_id, log_dir):
        logger = logging.getLogger("setu")
        logger.setLevel(logging.DEBUG)
        # ch = logging.StreamHandler(sys.stdout)
        # ch.setLevel(logging.INFO)
        fh = logging.FileHandler(log_dir + "/arjuna-setu-{}-ts-{}.log".format(time.time(), testsession_id), "w", 'utf-8')
        fh.setLevel(logging.DEBUG)
        f_fmt = logging.Formatter(u'[%(levelname)5s]\t%(asctime)s\t%(pathname)s::%(module)s.%(funcName)s:%(lineno)d\t%(message)s')
        c_fmt = logging.Formatter(u'[%(levelname)5s]\t%(message)s')
        # ch.setFormatter(c_fmt)
        fh.setFormatter(f_fmt)
        # logger.addHandler(ch)
        logger.addHandler(fh)
        cls.LOGGER = logger

    @classmethod
    def get_logger(cls):
        return cls.LOGGER and cls.LOGGER or cls.__DUMMY_LOGGER


__all__ = ["Setu", "GuiAutomator", "GuiElementMetaData", "SimpleGuiElementMetaData"]


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import importlib

import xml.etree.ElementTree as ETree

from arjuna.tpi import Arjuna
from arjuna.lib.utils import sys_utils
from arjuna.lib.utils import etree_utils
from arjuna.unitee.utils import run_conf_utils
from arjuna.unitee.types.root import Root
from arjuna.unitee.types.containers import *
from arjuna.lib.enums import *

from arjuna.lib.reader.textfile import *
from arjuna.lib.reader.hocon import *
from arjuna.unitee.test.defs.stage import *
from arjuna.unitee.test.objects.session import *
from arjuna.unitee.test.defs.fixture import *
from arjuna.lib.config import ConfigContainer


class __SessionDef(Root, metaclass=abc.ABCMeta):
    def __init__(self, sname):
        super().__init__()
        self.name = sname
        self.config = ConfigContainer()
        self.fpath = None
        self.raw_contents = None
        self.root = None
        self.stage_defs = []
        self.tmcount = 0
        # self.evars.update(self.central_config.clone_evars())
        self.__iter = None
        self.__fixtures = FixturesDef()
        self.logger = Arjuna.get_logger()
        self.unitee = Arjuna.get_unitee_instance()

    @property
    def fixture_defs(self):
        return self.__fixtures

    def process(self):
        def display_err_and_exit(msg):
            self.console.display_error((msg + " Fix session template file: {}").format(self.fpath))
            sys_utils.fexit()

        try:
            self.root = ETree.fromstring(self.raw_contents)
        except Exception as e:
            print(e)
            display_err_and_exit("Session definition could not be loaded because of errors in XML.")

        if self.root.tag != 'session':
            display_err_and_exit("Invalid session template file. Root element tag should be session.")

        node_dict = etree_utils.convert_to_cidict(self.root)

        if 'stages' not in node_dict:
            display_err_and_exit(">>stages<< must be specified in a session template.")

        for child_tag, child in node_dict.items():
            child_tag = child_tag.lower()
            if child_tag == 'config':
                config = child
                for option in config:
                    run_conf_utils.validate_config_xml_child("session", self.fpath, option)
                    run_conf_utils.add_config_node_to_configuration("session", self.config, option)
            elif child_tag == 'fixtures':
                fixtures = child
                for child in fixtures:
                    run_conf_utils.validate_fixture_xml_child("session", "session", self.fpath, child)
                    run_conf_utils.add_fixture_node_to_fixdefs(self.fixture_defs, child)
            elif child_tag =='stages':
                if "stage" not in etree_utils.convert_to_cidict(child):
                    display_err_and_exit(">>stages<< element in session definition must contain atleast one >>stage<< element.")

                stages = list(child)
                for index, stage in enumerate(stages):
                    run_conf_utils.validate_stage_xml_child("session", self.fpath, stage)
                    node = StageDef(self, index + 1, stage)
                    self.stage_defs.append(node)
            else:
                display_err_and_exit("Unexpected element >>{}<< found in session definition.".format(child.tag))

    def schedule(self):
        self.logger.debug("%s: Scheduling nodes".format(self.name))
        for stage in self.stage_defs:
            self.logger.debug("Session node: " + stage.get_name())
            stage.schedule()
            self.tmcount += stage.get_tmcount()
        self.__iter = iter(self.stage_defs)

    def pick(self):
        for stage_def in self.stage_defs:
            stage_def.pick()
        self.unitee.testdb.process_unpicked_and_skipped()
        self.unitee.testdb.process_dependencies()

        return TestSession(self)


class __MSessionDef(__SessionDef):
    def __init__(self, gname):
        super().__init__("msession")
        sr = TextResourceReader("st/msession.xml")
        contents = sr.read()
        sr.close()
        contents = contents.format(gname=gname)
        self.fpath = os.path.join(self.central_config.get_arjuna_option_value(ArjunaOption.ARJUNA_ROOT_DIR).as_string(),
                                  "arjuna/lib/res/st/msession.xml")
        self.raw_contents = contents


class MSessionAllTests(__MSessionDef):
    def __init__(self):
        super().__init__("_magroup")

class MSessionSingleGroup(__MSessionDef):
    def __init__(self, group_name):
        super().__init__(group_name)

class UserDefinedSessionDef(__SessionDef):
    def __init__(self, name, fpath):
        super().__init__(name)
        self.fpath = fpath
        f = open(self.fpath, "r")
        self.raw_contents = f.read()
        f.close()



import uuid
from arjuna.setuext.guiauto.impl.gui.namestore import GuiNameStore
from arjuna.setuext.guiauto.impl.gui.gui import Gui
from arjuna.tpi.enums import ArjunaOption
from arjuna.client.core.action import *
from .action_map import *
from arjuna.setu.adapter import Handler

class GuiHandlerManager:

    def __init__(self, project_config):
        self.__name_store = GuiNameStore()
        self.__namespace_dir = project_config.setu_config.value(ArjunaOption.GUIAUTO_NAMESPACE_DIR)
        self.__gui_map = {}

    @property
    def name_store(self):
        return self.__name_store

    @property
    def namespace_dir(self):
        return self.__namespace_dir

    def add_to_gui_map(self, setu_id, gui_handler):
        self.__gui_map[setu_id] = gui_handler

    def take_action(self, automator_handler, action_type, json_args):
        if action_type == GuiActionType.CREATE_GUI:
            gui_handler = self.create_gui(automator_handler, json_args)
            return {"guiSetuId" : gui_handler.setu_id}
        elif action_type == GuiActionType.CREATE_CHILD_GUI:
            return self.create_child_gui(automator_handler, json_args)
        else:
            raise Exception("Unknown action got for Gui Mananger: " + action_type.name)

    def create_gui(self, automator_handler, json_args):
        label = json_args["label"]
        defPath = json_args["defFileName"]
        gui = Gui(self.__name_store, self.__namespace_dir, automator_handler.automator, label, defPath)
        gui_handler = GuiHandler(self, automator_handler, gui)
        self.add_to_gui_map(gui.setu_id, gui_handler)
        return gui_handler

    def create_child_gui(self, automator_handler, json_args):
        parent_gui_setu_id = json_args.pop("parentGuiSetuId")
        parent_gui_handler = self.get_gui_handler(parent_gui_setu_id)
        return parent_gui_handler.create_gui(automator_handler, **json_args)

    def get_gui_handler(self, setu_id):
        return self.__gui_map[setu_id]


# Arg names of methods show JSON names, so don't follow Python conventions.
class GuiHandler:

    def __init__(self, gui_mgr, automator_handler, gui):
        self.__guimgr = gui_mgr
        self.__gui = gui
        self.__automator_handler = automator_handler
        self.__automator = None

    @property
    def gui(self):
        return self.__gui

    @property
    def setu_id(self):
        return self.__gui.setu_id

    @property
    def automator_handler(self):
        return self.__automator_handler

    def create_gui(self, automator_handler, label=None, name=None, qualName=None, defFileName=None):
        gui = Gui(self.__guimgr.name_store, self.__guimgr.namespace_dir, automator_handler.automator, label, defFileName)
        gui_handler = GuiHandler(self.__guimgr, automator_handler, gui)
        self.__guimgr.add_to_gui_map(gui.setu_id, gui_handler)
        return {"guiSetuId": gui.setu_id}

    def take_action(self, action_type, json_args):
        action_id = HANDLER_MAP[action_type]
        method_name, replaceable = HANDLER_NAME_MAP[action_id]
        return getattr(self, method_name)(action_type.name.replace(replaceable, "").lower(), json_args)

    def take_direct_action(self, action, json_args):
        Handler._pop_arg(json_args, "automatorSetuId")
        return getattr(self, action)(**json_args)

    def define_element(self, locators):
        emd = self.gui.get_emd(locators)
        return self.automator_handler.define_element_with_emd(emd)

    def define_multielement(self, locators):
        emd = self.gui.get_emd(locators)
        return self.automator_handler.define_multielement_with_emd(emd)

    def define_dropdown(self, locators):
        emd = self.gui.get_emd(locators)
        return self.automator_handler.define_dropdown_with_emd(emd)

    def define_radiogroup(self, locators):
        emd = self.gui.get_emd(locators)
        return self.automator_handler.define_radiogroup_with_emd(emd)

    def define_frame(self, locators):
        emd = self.gui.get_emd(locators)
        return self.automator_handler.define_frame_with_emd(emd)

    def define_alert(self):
        return self.automator_handler.define_alert()

    def define_main_window(self):
        return self.automator_handler.get_main_window()

    def set_slomo(self, on, interval=None):
        self.automator_handler.set_slomo(on, interval)

    def take_window_action(self, action, json_dict):
        return self.automator_handler.take_window_action(action, json_dict)

    def take_main_window_action(self, action, json_dict):
        return self.automator_handler.take_main_window_action(action, json_dict)

    def take_child_window_action(self, action, json_dict):
        return self.automator_handler.take_child_window_action(action, json_dict)

    def take_browser_action(self, action, json_dict):
        return self.automator_handler.take_browser_action(action, json_dict)

    def take_element_action(self, action, json_dict):
        return self.automator_handler.take_element_action(action, json_dict)

    def take_multielement_action(self, action, json_dict):
        return self.automator_handler.take_multielement_action(action, json_dict)

    def take_dropdown_action(self, action, json_dict):
        return self.automator_handler.take_dropdown_action(action, json_dict)

    def take_radiogroup_action(self, action, json_dict):
        return self.automator_handler.take_radiogroup_action(action, json_dict)

    def take_alert_action(self, action, json_dict):
        return self.automator_handler.take_alert_action(action, json_dict)

    def take_domroot_action(self, action, json_dict):
        return self.automator_handler.take_domroot_action(action, json_dict)

    def take_frame_action(self, action, json_dict):
        return self.automator_handler.take_frame_action(action, json_dict)


import abc

class Handler(metaclass=abc.ABCMeta):

    @classmethod
    def _pop_arg(cls, json_args_dict, key, optional=False):
        try:
            value = json_args_dict[key]
            del json_args_dict[key]
            return value
        except Exception as e:
            print(e)
            if optional: 
                return None
            else: 
                raise Exception("Input value for {} not found in JSON args: {}.".format(key, json_args_dict))

    @classmethod
    def get_element_setuid(cls, json_dict):
        return cls._pop_arg(json_dict, "elementSetuId")

    @classmethod
    def get_config_setuid(cls, json_dict):
        return cls._pop_arg(json_dict, "configSetuId")

    @classmethod
    def get_extended_config(cls, json_dict):
        return cls._pop_arg(json_dict, "extendedConfig", optional=True)


import inspect
from arjuna.tpi.enums import ArjunaOption
from .driverelement import SeleniumDriverElement


class SeleniumDriver:

    def __init__(self, setu_id, requester):
        self.__setu_id = setu_id
        self.__requester = requester

    def create_gui_element_dispatcher(self, element_setu_id):
        return SeleniumDriverElement()
        self.__dispatcher.create_gui_element_dispatcher(element_setu_id)

    def __post_to_actor(self, action, actor_uri, **kwargs):
        input_dict = {"automatorSetuId" : self.setu_id}
        input_dict.update(kwargs)
        json_dict = {
            "action" : action,
            "args" : input_dict
        }
        return self.__requester.post(actor_uri, json_dict)

    def post_action(self, **kwargs):
        action = inspect.stack()[1].function
        return self.__post_to_actor(action, "/automator/action", **kwargs)

    def post_other(self, url, **kwargs):
        action = inspect.stack()[1].function
        return self.__post_to_actor(action, url, **kwargs)

    @property
    def setu_id(self):
        return self.__setu_id

    def launch(self, config):
        self.post_other("/automator/launch", 
        automatorName=config["arjunaOptions"][ArjunaOption.GUIAUTO_AUTOMATOR_NAME.name], 
        config=config)

    def quit(self):
        self.post_action()

    def go_to_url(self, url):
        self.post_action(url=url)

    def go_back_in_browser(self):
        self.post_action()

    def go_forward_in_browser(self):
        self.post_action()

    def refersh_browser(self):
        self.post_action()

    def execute_javascript(self, script):
        self.post_action(script=script)

    def take_screenshot(self):
        self.post_action()

    def find_element(self, child_gui_element_set_id, with_type, with_value):
        self.post_action(
            elementSetuId = child_gui_element_set_id,
            withType = with_type,
            withValue = with_value
        )

    def find_multielement(self, child_gui_element_set_id, with_type, with_value):
        response = self.post_action(
            elementSetuId = child_gui_element_set_id,
            withType = with_type,
            withValue = with_value
        )

        return response["data"]["instanceCount"]

    def __get_value_from_response(self, response, key):
        if "data" not in response:
            raise Exception("Setu actor json response: {} does not contain 'data' key.")
        elif key not in response:
            raise Exception("Setu actor json response: {} does not contain key <{}> in data section.")
        else:
            return response["data"][key]

    def __get_result(self, response):
        return self.__get_value_from_response(response, "result")

    def get_current_window_handle(self):
        response = self.post_action()
        return self.__get_result(response)

    def get_current_window_title(self):
        response = self.post_action()
        return self.__get_result(response)

    def maximize_current_window(self):
        self.post_action()

    def get_current_window_size(self):
        response = self.post_action()
        size = self.__get_result(response)
        return size["width"], size["height"]

    def get_all_window_handles(self):
        response = self.post_action()
        return self.__get_result(response)

    def focus_on_window(self, handle):
        self.post_action(handle=handle)

    def close_current_window(self):
        self.post_action()

    def is_web_alert_present(self):
        response = self.post_action()
        return self.__get_result(response)

    def confirm_web_alert(self):
        self.post_action()

    def dismiss_web_alert(self):
        self.post_action()

    def get_text_from_web_alert(self):
        response = self.post_action()
        return self.__get_result(response)

    def send_text_to_web_alert(self,text):
        self.post_action(text=text)

    def get_current_mobile_view_context(self):
        response = self.post_action()
        return self.__get_result(response)

    def get_all_mobile_view_contexts(self):
        response = self.post_action()
        return self.__get_result(response)

    def focus_on_mobile_view_context(self, view_context):
        self.post_action(viewContext=view_context)

    def focus_on_frame(self, elem_setu_id, is_instance_action=False, instance_index=0):
        if is_instance_action:
            self.post_action(elementSetuId=elem_setu_id, isInstanceAction=is_instance_action, instanceIndex=instance_index)
        else:
            self.post_action(elementSetuId=elem_setu_id)

    def focus_on_parent_frame(self):
        self.post_action()

    def focus_on_dom_root(self):
        self.post_action()



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.unitee.loader.mod_loader import *
from arjuna.lib.discovery import *
from arjuna.lib.utils import obj_utils
from arjuna.tpi.enums import ArjunaOption


class TestLoader:
    def __init__(self):
        self.lock = threading.RLock()
        self.__loading_completed = False
        self.__mlmap = {}
        from arjuna.tpi import Arjuna
        central_config = Arjuna.get_central_config()
        self.proj_dir = central_config.get_arjuna_option_value(ArjunaOption.PROJECT_ROOT_DIR).as_string()
        self.test_dir = central_config.get_arjuna_option_value(ArjunaOption.UNITEE_PROJECT_TESTS_DIR).as_string()
        self.logger = Arjuna.get_logger()
        self.console = Arjuna.get_console()

    def load(self):
        self.logger.debug("Now finding tests inside: " + self.proj_dir)
        aggregator = FileAggregator()
        discoverer = FileDiscoverer(aggregator, self.proj_dir, [self.test_dir + "/" + "modules"])
        discoverer.discover()
        aggregator.enumerate()
        for def_file in aggregator:
            file_name = def_file.attr(DiscoveredFileAttributeEnum.FULL_NAME)
            # Arjuna would not import module names that begin with _
            if file_name.startswith("_"): continue
            self.logger.debug("Found: " + def_file.attr(DiscoveredFileAttributeEnum.FULL_NAME))
            self.__handle_discovered_module(def_file)

        ### This STEP needs to be looked into AFTER complete processing is working.
        ###Unitee.def_db.register_with_reporter()

    def __get_pkg_module_qname_for_discovered_file(self, f):
        qname = None
        from arjuna.tpi import Arjuna
        project = Arjuna.get_central_arjuna_option(ArjunaOption.PROJECT_NAME).as_string()
        pkg = ".".join([project, f.attr(DiscoveredFileAttributeEnum.PACKAGE_DOT_NOTATION).strip()])
        module = f.attr(DiscoveredFileAttributeEnum.NAME).strip()
        qname = ".".join([pkg, module])
        return pkg, module, qname

    def __get_pkg_module_qname_from_mqname(self, mqname):
        pkg, module = mqname.rsplit(".", 1)
        qname = mqname
        return pkg, module, mqname

    def __handle_discovered_module(self, def_file):
        pkg, module, qname = self.__get_pkg_module_qname_for_discovered_file(def_file)
        if def_file.attr(DiscoveredFileAttributeEnum.EXTENSION).lower() != "py":
            self.logger.debug("Non-Python file or non-py extension Python file: " + qname)

        if qname not in self.__mlmap:
            self.logger.debug("Loading module: " + qname)
            self.__mlmap[qname] = ModuleLoader(pkg, module, qname)
            self.__mlmap[qname].load()
        else:
            self.logger.debug("Module was already loaded via test script import: " + qname)

        # Here java took an annotation based decision whether it is a test class
        # This should take place now in Module Loader. If @tdmodule is not marked
        # or no tests are present
        # Also, a non-test class should be added by Module Loader to non-test class queue
        # of Unitee.def_db

    def __handle_indirect_imported_module(self, mqname):
        pkg, module, qname = self.__get_pkg_module_qname_from_mqname(mqname)
        self.logger.debug("Loading module: " + qname)
        self.__mlmap[qname] = ModuleLoader(pkg, module, qname)
        self.__mlmap[qname].load()

    def freeze(self):
        self.__loading_completed = True

    def __validate_state(self, kallable, dec_type):
        if self.__loading_completed:
            ktype = type(kallable) is type and "class" or "function"
            from arjuna.tpi import Arjuna
            Arjuna.get_console().display_error(
                "You are decorating inner {} {} with {} in {} module. These decorators are ignored by Arjuna.".format(
                    ktype,
                    kallable.__qualname__,
                    dec_type,
                    kallable.__module__
                ))
            return False
        return True

    def __register(self, dec, what, kallable, *vars, **kwargs):
        proceed = self.__validate_state(kallable, "@{}".format(dec))
        if proceed:
            ml = self.__get_ml(kallable)
            return getattr(ml, "register_" + what)(kallable, *vars, **kwargs)

    def __get_ml(self, kallable):
        if obj_utils.callable(kallable):
            if kallable.__module__ not in self.__mlmap:
                self.__handle_indirect_imported_module(kallable.__module__)
            return self.__mlmap[kallable.__module__]
        else:
            return kallable

    @sync_method('lock')
    def register_tfunc(self, kallable, **tsargs):
        return self.__register("tfunc", "tfunc", kallable, **tsargs)
        # proceed = self.__validate_state(kallable, "@test_source or @ts")
        # if proceed:
        #     ml = self.__get_ml(kallable)
        #     ml.register_tfunc(kallable, **tsargs)
        # return kallable

    @sync_method('lock')
    def register_tmodule(self, kallable, **tsargs):
        proceed = self.__validate_state(kallable, "@test_source or @ts")
        if proceed:
            ml = self.__get_ml(kallable)
            ml.register_tmodule(kallable, **tsargs)
        return kallable

    @sync_method('lock')
    def register_fixture(self, ftype, kallable):
        proceed = self.__validate_state(kallable, "@{}".format(ftype))
        if proceed:
            ml = self.__get_ml(kallable)
            ml.register_fixture(ftype, kallable)
        return kallable

    @sync_method('lock')
    def register_skip_func(self, kallable):
        proceed = self.__validate_state(kallable, "@skip_me")
        if proceed:
            ml = self.__get_ml(kallable)
            ml.register_skip(kallable)
        return kallable

    def enumerate(self):
        names = list(self.__mlmap.keys())
        names.sort()
        for name in names:
            ml = self.__mlmap[name]
            ml.enumerate()


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import os

from arjuna.lib.enums import *
from arjuna.lib.utils import file_utils
from arjuna.lib.utils import sys_utils

class DiscoveredFile:
    def __init__(self):
        self.props = {}
        
    def attr(self, attr):
        return self.props[attr]

    def set_attr(self, name, attr):
        self.props[name] = attr


fa_msg1 = '''Duplicate test file found with name: %s.
Check package and class names in test directory.
Arjuna follows a case-INSENSITIVE approach for test names.'''


class FileAggregator:
    def __init__(self):
        self.files = []
        self.found_class_names = set()
        self.temp_map = {}
        from arjuna.tpi import Arjuna
        self.logger = Arjuna.get_logger()
        self.console = Arjuna.get_console()

    def add(self, df):
        key = df.attr(DiscoveredFileAttributeEnum.DIRECTORY_ABSOLUTE_PATH) + "/" + df.attr(
            DiscoveredFileAttributeEnum.FULL_NAME).upper()
        if key in self.found_class_names or key in self.temp_map:
            fa_msg1.format(DiscoveredFileAttributeEnum.FULL_NAME)
            self.console.display_error(fa_msg1)
            sys_utils.fexit()
        self.temp_map[key] = df

    def freeze(self):
        paths = []
        paths.extend(self.temp_map.keys())
        paths.sort()
        for path in paths:
            self.files.append(self.temp_map[path])

    def __iter__(self):
        return iter(self.files)

    def enumerate(self):
        for f in self:
            self.logger.debug("-------------------------")
            self.logger.debug("Name:\t" + f.attr(DiscoveredFileAttributeEnum.NAME))
            self.logger.debug("Extension:\t" + f.attr(DiscoveredFileAttributeEnum.EXTENSION))
            self.logger.debug("Full Name:\t" + f.attr(DiscoveredFileAttributeEnum.FULL_NAME))
            self.logger.debug("Package Dot Notation:\t" + f.attr(DiscoveredFileAttributeEnum.PACKAGE_DOT_NOTATION))
            self.logger.debug(
                "Directory Relative Path:\t" + f.attr(DiscoveredFileAttributeEnum.DIRECTORY_RELATIVE_PATH))
            self.logger.debug(
                "Directory Absolute Path:\t" + f.attr(DiscoveredFileAttributeEnum.DIRECTORY_ABSOLUTE_PATH))
            self.logger.debug("Comma Separated Relative Path:\t"
                         + f.attr(DiscoveredFileAttributeEnum.COMMA_SEPATARED_RELATIVE_PATH))
            # self.logger.debug("Container:\t" + f.attr(DiscoveredFileAttributeEnum.CONTAINER))
            # self.logger.debug("Container Type:\t" + f.attr(DiscoveredFileAttributeEnum.CONTAINER_TYPE))
            self.logger.debug("-------------------------")


class FileDiscoverer:
    def __init__(self, aggregator, root_dir, include_prefixes=None):
        self.aggregator = aggregator
        if root_dir.endswith("\\") or root_dir.endswith("//"):
            self.root_dir = root_dir[0:-1]
        else:
            self.root_dir = root_dir
        self.root_dir = file_utils.normalize_path(self.root_dir)
        self.cdir = None
        self.cabsdir = None
        self.prefixes = [file_utils.normalize_path(p) for p in include_prefixes]

    def discover(self):
        for d, subdlist, flist in os.walk(self.root_dir):
            normalized_d = file_utils.normalize_path(d)
            if flist:
                for f in flist:
                    full_path = file_utils.normalize_path(os.path.abspath(os.path.join(normalized_d, f)))
                    consider = False
                    for prefix in self.prefixes:
                        if normalized_d.startswith(prefix):
                            consider = True
                            break
                    if not consider: continue
                    file_ext = file_utils.get_extension(full_path)
                    if file_ext.lower() not in set(['py']): continue
                    parent_dir = file_utils.normalize_path(os.path.dirname(full_path))
                    pkg_parent_dir = None  # os.path.commonpath()
                    if parent_dir == self.root_dir:
                        pkg_parent_dir = ""
                    else:
                        pkg_parent_dir = parent_dir[parent_dir.index(self.root_dir) + len(self.root_dir) + 1:]
                    file_ext = file_utils.get_extension(full_path)

                    df = DiscoveredFile()
                    df.set_attr(DiscoveredFileAttributeEnum.NAME, file_utils.get_nonext_basename(full_path))
                    df.set_attr(DiscoveredFileAttributeEnum.EXTENSION, file_ext)
                    df.set_attr(DiscoveredFileAttributeEnum.FULL_NAME, f)
                    df.set_attr(DiscoveredFileAttributeEnum.PACKAGE_DOT_NOTATION,
                                 pkg_parent_dir.replace("/", "."))
                    df.set_attr(DiscoveredFileAttributeEnum.DIRECTORY_RELATIVE_PATH, pkg_parent_dir)
                    df.set_attr(DiscoveredFileAttributeEnum.DIRECTORY_ABSOLUTE_PATH, parent_dir)
                    replaced = df.attr(DiscoveredFileAttributeEnum.DIRECTORY_RELATIVE_PATH).replace("/", "|")
                    replaced = replaced.replace("\\", "|")
                    df.set_attr(DiscoveredFileAttributeEnum.COMMA_SEPATARED_RELATIVE_PATH,
                                 ",".join(replaced.split("|")))
                    # df.set_attr(DiscoveredFileAttributeEnum.CONTAINER, attr.NA_STRING)
                    # df.set_attr(DiscoveredFileAttributeEnum.CONTAINER_TYPE, attr.NA_STRING)
                    self.aggregator.add(df)

        self.aggregator.freeze()



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.unitee.loader.deptree import DepTreeBuilder

class DefDB:
    def __init__(self):
        self.__dep_tree_builder = DepTreeBuilder()
        # For pulling out class definitions by name
        self.__skipped = []
        self.__unpicked = []

        self.__mdefs = {}
        self.__mqueue = []
        # The following gets populated from above, if classDef.isNotPickedByAnyGroup() is True
        self.__for_processor = []
        from arjuna.tpi import Arjuna
        self.logger = Arjuna.get_logger()

    def register_mdef(self, mdef):
        qname = mdef.qname
        self.logger.debug("Registering method definition for: {}".format(qname))
        self.__mdefs[qname] = mdef
        self.__mqueue.append(qname)

    def freeze(self):
        self.__iter = iter(self.__mdefs)

    def get_mnames(self):
        return self.__mqueue

    def get_mdef(self, mname):
        return self.__mdefs[mname]

    def __iter__(self):
        return iter(self.__mdefs.items())

    def __should_include(self, fname, unpicked_set, skip_set):
        if fname in unpicked_set:
            return False
        if fname in skip_set:
            return False
        return True

    def process_unpicked_and_skipped(self):
        temp1 = set()
        temp2 = set()
        for mname, mdef in self:
            if mdef.unpicked:
                self.__unpicked.append((mdef.pkg, mdef.module))
                temp1.add(mname)
            elif mdef.skipped:
                self.__skipped.append((mdef.pkg, mdef.module))
                temp2.add(mname)
            else:
                mdef.process_unpicked_and_skipped()
        self.__mdefs = {i:j for i,j in self.__mdefs.items() if self.__should_include(i, temp1, temp2)}
        self.__mqueue = [i for i in self.__mqueue if i in self.__mdefs]

    def process_dependencies(self):
        for name in self.__mqueue:
            mdef = self.__mdefs[name]
            mdef.process_dependencies()
            if mdef.dependency:
                mdef.dependency.process()
                for du in mdef.dependency:
                    if du.name not in self.__mdefs:
                        du.set_ignored()
            self.__dep_tree_builder.create_node(name, mdef.dependency)

        self.__dep_tree_builder.process_dependencies()
        self.__dep_tree_builder.validate()


    def slot_module_names(self, module_names):
        return self.__dep_tree_builder.slot_names(module_names)

    def get_mdef_queue_for_def_processor(self):
        return self.__for_processor


    def get_tsclass_definition(name):
        return test_class_definitions[name]

    def get_dependency_tree_builder(self):
        return self.__dep_tree_builder


    def is_test_class(self, name):
        return name in self._mdefs


    def is_test_class_marked_as_skipped(self, name):
        if not is_test_class(name):
            raise Exception("Not a test class")
        else:
            return self._mdefs[name].skipped


    def has_class(self, name):
        return name in self.__all_class_name_set


    def should_run_module(self, qname):
        return qname in self.__mqueue


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.tpi import Arjuna
from arjuna.unitee.types.root import *
from arjuna.unitee.types.containers import *
from arjuna.lib.reader.hocon import *
from arjuna.lib.utils import sys_utils
from arjuna.unitee.test.defs.fixture import *

from .group import *

from xml.etree.ElementTree import Element

from arjuna.lib.utils import etree_utils
from arjuna.unitee.utils import run_conf_utils
from arjuna.lib.config import ConfigContainer

class StageDef(Root):
    def __init__(self, sdef, id, stage_xml):
        super().__init__()
        self.gdefs = []
        self.config = ConfigContainer()
        self.tmcount = 0
        self.threads = 1
        self.sdef = sdef
        self.__iter = None
        self.__fixtures = FixturesDef()
        self.root = stage_xml

        self.unitee = Arjuna.get_unitee_instance()

        self.id = id
        self.name = "stage{:d}".format(id)

        if not isinstance(stage_xml, Element):
            self.console.display_error("Fatal: [Arjuna Error] Unsuppored input argument supplied for stage creation: {}".format(stage_xml))
            sys_utils.fexit()
        else:
            self.__process(stage_xml)
            # self.nodes.append(SessionSubNode(self, len(self.nodes) + 1, input))

    @property
    def fixture_defs(self):
        return self.__fixtures

    def __process(self, group_hocon):
        def display_err_and_exit(msg):
            self.console.display_error((msg + " Fix session template file: {}").format(self.sdef.fpath))
            sys_utils.fexit()

        stage_attrs = etree_utils.convert_attribs_to_cidict(self.root)

        if "name" in stage_attrs:
            self.name = stage_attrs['name'].strip()
            if not self.name:
                display_err_and_exit(">>name<< attribute in stage definition should be a non-empty string.")

        threads_err_msg = ">>threads<< attribute in stage definition can be integer >=1."
        if "threads" in stage_attrs:
            self.threads = stage_attrs['threads'].strip()
            try:
                self.threads = int(self.threads)
            except:
                display_err_and_exit(threads_err_msg)
            else:
                if self.threads <=0:
                    display_err_and_exit(threads_err_msg)


        node_dict = etree_utils.convert_to_cidict(self.root)

        if "groups" not in node_dict:
            display_err_and_exit(">>stage<< element in session definition must contain >>groups<< element.")

        for child_tag, child in node_dict.items():
            child_tag = child_tag.lower()
            if child_tag == 'config':
                config = child
                for option in config:
                    run_conf_utils.validate_config_xml_child("session", self.sdef.fpath, option)
                    run_conf_utils.add_config_node_to_configuration("session", self.config, option)
            elif child_tag == 'fixtures':
                fixtures = child
                for child in fixtures:
                    run_conf_utils.validate_fixture_xml_child("session", "stage", self.sdef.fpath, child)
                    run_conf_utils.add_fixture_node_to_fixdefs(self.fixture_defs, child)
            elif child_tag =='groups':
                if "group" not in etree_utils.convert_to_cidict(child):
                    display_err_and_exit(">>groups<< element in stage definition must contain atleast one >>group<< element.")

                groups = list(child)

                for index, group in enumerate(groups):
                    run_conf_utils.validate_group_xml_child("session", self.sdef.fpath, group)
                    node = GroupDef(self.sdef, self, len(self.gdefs) + 1, group)
                    self.gdefs.append(node)
            else:
                display_err_and_exit("Unexpected element >>{}<< found in >>stage<< definition in session file.".format(child.tag))

    def pick(self):
        for gdef in self.gdefs:
            self.tmcount += gdef.pick()

        self.__iter = iter(self.gdefs)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import inspect
import types


def get_class_name(obj):
    o = isinstance(obj, type) and obj or type(obj)
    return o.__name__


def get_class_qual_name(obj):
    o = isinstance(obj, type) and obj or type(obj)
    return o.__module__ + "." + o.__name__


def callable(obj):
    return not is_object(obj)


def is_module(obj):
    return type(obj) is types.ModuleType


def is_object(obj):
    return not isinstance(obj, type) and not type(obj) is types.FunctionType


def is_class(obj):
    return isinstance(obj, type)


def is_function(obj):
    return type(obj) is types.FunctionType and obj.__qualname__ == obj.__name__


def is_method(obj):
    return type(obj) is types.FunctionType and obj.__qualname__ != obj.__name__


def is_private(obj):
    return obj.__name__.startswith('__')


def is_protected(obj):
    return obj.__name__.startswith('_')


def is_public(obj):
    return not is_private(obj) and not is_protected(obj)


class IterItemOp:
    def __init__(self, i):
        self._i = i

    def __eq__(self, obj):
        if type(obj) is bool:
            for i in self._i:
                if bool(i) != obj:
                    return False
            return True
        else:
            for i in self._i:
                if i != obj:
                    return False
            return True


def is_my_sig(obj):
    args, *rest = inspect.getfullargspec(obj)
    if args == ['my'] and IterItemOp(rest) == False:
        return True
    else:
        return False



from arjuna.client.core.connector import SetuArg
from arjuna.lib.config import DefaultTestConfig
from arjuna.client.core.connector import BaseSetuObject
from arjuna.client.core.action import *
from arjuna.client.core.config import ArjunaComponent

class DefaultTestSession(BaseSetuObject):
    
    def __init__(self):
        super().__init__()
        self.__DEF_CONF_NAME = "central"

    def init(self, project_root_dir, cli_config, runid):
        super().__init__()
        args = [SetuArg.arg("projectRootDir", project_root_dir)]
        if cli_config:
            args.append(SetuArg.arg("cliConfig", cli_config.as_map()))
        if runid:
            args.append(SetuArg.arg("runId", runid))

        response = self._send_request(
            ArjunaComponent.TEST_SESSION,
            TestSessionActionType.INIT,
            *args
        )
        self._set_setu_id(response.get_value_for_testsession_setu_id())
        self._set_self_setu_id_arg("testSessionSetuId")
        return self.__create_config_from_response(response)

    def __create_config_from_response(self, response, name=None):
        res_data = response.get_data()
        config = DefaultTestConfig(
            self,
            name and name or self.__DEF_CONF_NAME,
            response.get_value_for_config_setu_id(),
            res_data["arjunaOptions"],
            res_data["userOptions"]
        )
        return config

    def finish(self):
        pass
        # To do

    def __register_config(self, name, hasParent, parentConfigId, arjunaOptions, userOptions):
        response = self._send_request(
                ArjunaComponent.CONFIGURATOR,
                ConfigActionType.REGISTER_NEW_CONFIG,
                SetuArg.arg("hasParent", hasParent),
                SetuArg.arg("parentConfigId", parentConfigId),
                SetuArg.arg("arjunaOptions", arjunaOptions),
                SetuArg.arg("userOptions", userOptions)
        )
        return self.__create_config_from_response(response, name)

    def register_config(self, name, arjuna_options, user_options):
        return self.__register_config(name, False, None, arjuna_options, user_options)

    def register_child_config(self, name, parent_conf_id, arjuna_options, user_options):
        return self.__register_config(name, True, parent_conf_id, arjuna_options, user_options)

    def create_file_data_source(self, record_type, file_name, *arg_pairs):
        response = self._send_request(
            ArjunaComponent.DATA_SOURCE,
            DataSourceActionType.CREATE_FILE_DATA_SOURCE,
            *arg_pairs
        )
        return response.get_data_source_id()

    def create_gui(self, automator, *setu_args):
        args = setu_args + (SetuArg.arg("automatorSetuId", automator.get_setu_id()), )
        response = self._send_request(
            ArjunaComponent.GUI,
            GuiActionType.CREATE_GUI,
            *args
        )
        return response.get_gui_setu_id()








from arjuna.unitee.enums import *

built_in_prop_type = {
    BuiltInProp.ID : str,
    BuiltInProp.PRIORITY : int,
    BuiltInProp.THREADS : int,
    BuiltInProp.NAME : str,
    BuiltInProp.AUTHOR : str,
    BuiltInProp.IDEA : str,
    BuiltInProp.UNSTABLE : bool,
    BuiltInProp.COMPONENT : str,
    BuiltInProp.APP_VERSION : str,
}

def validate_built_in_props(props):
    for k,v in props.items():
        if is_builtin_prop(k.upper()):
            expected_type = built_in_prop_type[BuiltInProp[k.upper()]]
            actual_type = type(v)
            if v is not None and actual_type is not expected_type:
                raise Exception("Built-in property {} should of type {}. Found {} of type {}".format(
                    k,
                    expected_type,
                    v,
                    actual_type
                ))

def get_value_type(built_in_prop_name):
    return built_in_prop_type[BuiltInProp[built_in_prop_name.upper()]]

def is_builtin_prop(name):
    return name.upper() in BuiltInProp.__members__


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import abc

from arjuna.lib.reader.excel import *
from .record import *


class ContextualDataReference(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def record_for(self, context):
        pass

class __ExcelDataReference(ContextualDataReference):
    def __init__(self, path, key_column=None):
        self.map = {}
        self.path = path
        self.key_column = key_column
        if (path.lower().ends_with("xls")):
            self.reader = ExcelRow2ArrayReader(path)
        else:
            raise Exception("Unsupported file format for Excel reading.")

        self._populate()

    @abc.abstractmethod
    def _populate(self):
        pass

    def record_for(self, context):
        if context.lower() in self.map:
            return self.map[context.lower()]
        else:
            raise Exception("{} at {} does not contain {} context key.".format(self.__class__.__name__, self.path, context))

class ExcelRowDataReference(__ExcelDataReference):

    def __init__(self, path):
        super().__init__(path)

    def _populate(self):
        headers = self.reader.next()[1:]
        while True:
            try:
                data_record = self.reader.next()
            except StopIteration:
                break
            else:
                self.map[data_record[0].lower()] = DataRecord(zip(headers, data_record[1:]))
        self.reader.close()


class ExcelColumnDataReference(__ExcelDataReference):

    def __init__(self, path):
        super().__init__(path)

    def _populate(self):
        contexts = self.reader.next()[1:]
        self.map = {i: {} for i in contexts}
        while True:
            try:
                data_record = self.reader.next()
            except StopIteration:
                break
            else:
                header = data_record[0]
                for index, context in enumerate(contexts):
                    self.map[context][header] = data_record[index]
        self.reader.close()



'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.tpi import Arjuna
from arjuna.lib.thread.decorators import *
from arjuna.unitee.reporter.result import *
from arjuna.unitee.exceptions import *
from arjuna.lib.exceptions import *
from arjuna.lib.utils import thread_utils
from arjuna.unitee.enums import *
from arjuna.unitee.validation.testpoint import *

import traceback

from arjuna.unitee.enums import *

class _SummaryResult:
    def __init__(self):
        self.summary = {i: 0 for i in ResultTypeEnum}

    def increment(self, type):
        self.summary[type] += 1

    def succeeded(self):
        return self.summary[ResultTypeEnum.PASS] > 0 \
               and self.summary[ResultTypeEnum.FAIL] == 0 \
               and self.summary.get[ResultTypeEnum.ERROR] == 0\
                and self.summary.get[ResultTypeEnum.EXCLUDED] == 0

    def has_issues(self):
        return not self.succeeded()

    def has_failures(self):
        return self.summary[ResultTypeEnum.FAIL] > 0

    def has_errors(self):
        return self.summary[ResultTypeEnum.ERROR] > 0

    def has_exclusions(self):
        return self.summary[ResultTypeEnum.EXCLUDED] > 0 or self.summary[ResultTypeEnum.EXCLUDED_CHILDREN] > 0

    def update_from_summary_result(self, summary):
        for k,v in summary.summary.items():
            self.summary[k] = self.summary[k] + v

class _ThreadState:

    def __init__(self):
        self.lock = threading.RLock()
        self.logger = Arjuna.get_logger()
        self.__current_steps = None
        self.__current_context = None
        self.unitee = Arjuna.get_unitee_instance()

    def get_steps(self):
        return tuple(self.__current_steps)

    def begin_recording(self, context):
        self.__current_steps = []
        self.__current_context = context

    def end_recording(self):
        self.__current_context = None
        return self.__current_steps

    @sync_method('lock')
    def __add_step(self, step):
        step.exec_context = self.__current_context
        if self.__current_context == ExecutionContext.Test:
            step.otype = "TestStep"
        else:
            step.otype = "FixtureStep"
        self.__current_steps.append(step)

    @sync_method('lock')
    def __get_step_id(self):
        return len(self.__current_steps) + 1

    @sync_method('lock')
    def add_pstep(self, step):
        step.step_num = self.__get_step_id()
        self.__add_step(step)

    @sync_method('lock')
    def add_fstep(self, step):
        step.step_num = self.__get_step_id()
        issue = self.__create_issue_for_problem_step(ResultTypeEnum.FAIL)
        step.issue = issue
        step.iid = issue.iid
        self.__add_step(step)

    @sync_method('lock')
    def add_estep(self, step):
        step.step_num = self.__get_step_id()
        issue = self.__create_issue_for_problem_step(ResultTypeEnum.ERROR)
        step.issue = issue
        step.iid = issue.iid
        self.__add_step(step)

    @sync_method('lock')
    def add_step_exception(self, e, strace):
        step = self.__create_step_result_for_exception(self.__get_step_id(), e, strace)
        self.__add_step(step)

    def __create_step_result_for_exception(self, step_num, e, strace):
        try:
            if isinstance(e, Failure):
                return self.__create_step_failure(step_num, e, strace)
            elif isinstance(e, AssertionError):
                return self.__create_py_step_failure(step_num, e, strace)
            elif isinstance(e, Error):
                return self.__create_step_error(step_num, e, strace)
            elif isinstance(e, Problem):
                return self.__create_py_step_error(step_num, e, strace)
            else:
                return self.__create_py_step_error(step_num, e, strace)
        except Exception as f:
            return self.__create_py_step_error(step_num, f, traceback.format_exc())

    def __create_issue(self, e, strace, i_type):
        return self.unitee.state_mgr.create_issue_from_exception(e, strace, i_type)

    def __create_issue_for_problem_step(self, i_type):
        return self.unitee.state_mgr.create_issue_for_problem_step(i_type)

    def __create_step_failure(self, step_num, e, strace):
        issue = self.__create_issue(e, strace, ResultTypeEnum.FAIL)
        step = e.step
        step.rtype = issue.itype
        step.step_num = step_num
        step.issue = issue
        step.iid = issue.iid
        return step

    def __create_py_step_failure(self, step_num, e, strace):
        issue = self.__create_issue(e, strace, ResultTypeEnum.FAIL)
        step = Step()
        step.iid = issue.iid
        step.rtype = ResultTypeEnum.FAIL
        step.step_num = step_num
        step.issue = issue
        step.set_failure()
        return step

    def __create_step_error(self, step_num, e, strace):
        issue = self.__create_issue(e, strace, ResultTypeEnum.ERROR)
        step = e.step
        step.rtype = issue.itype
        step.step_num = step_num
        step.issue = issue
        step.iid = issue.iid
        return step

    def __create_py_step_error(self, step_num, e, strace):
        issue = self.__create_issue(e, strace, ResultTypeEnum.ERROR)
        step = Step()
        step.iid = issue.iid
        step.rtype = ResultTypeEnum.ERROR
        step.step_num = step_num
        step.issue = issue
        step.set_error()
        return step

class SummaryState:
    def __init__(self):
        self.logger = Arjuna.get_logger()
        self.summary_result = _SummaryResult()
        self.test_summary = _SummaryResult()

    def increment(self, rtype):
        self.summary_result.increment(rtype)

    def update(self, result):
        rtype =  ResultTypeEnum[result.rtype]
        self.increment(rtype)

    def update_test_summary_from_state(self, state):
        self.test_summary.update_from_summary_result(state.test_summary)

    def has_issues(self):
        return self.summary_result.has_issues()

    def has_failures(self):
        return self.summary_result.has_failures()

    def has_errors(self):
        return self.summary_result.has_errors()

    def has_exclusions(self):
        return self.summary_result.has_exclusions()

class StateWithIssueMgt(SummaryState):

    def __init__(self):
        super().__init__()
        self.__issues_map = {}

    def update(self, result):
        rtype =  ResultTypeEnum[result.rtype]
        self.increment(rtype)
        name = self._get_child_name(result)
        iid = result.iid

        if rtype in {ResultTypeEnum.FAIL, ResultTypeEnum.ERROR, ResultTypeEnum.EXCLUDED}:
            self.__issues_map[name] = iid

    def _get_child_name(self, result):
        pass

    def raise_on_problem(self, name):
        if name in self.__issues_map:
            raise DependencyNotMet(self.__issues_map[name])

SessionState = SummaryState
StageState = SummaryState
FunctionState = SummaryState

class GroupState(StateWithIssueMgt):
    def __init__(self):
        super().__init__()

    def _get_child_name(self, result):
        return ".".join([result.pkg, result.module])

class ModuleState(StateWithIssueMgt):
    def __init__(self):
        super().__init__()

    def _get_child_name(self, result):
        return result.function

class ThreadManager:
    def __init__(self):
        self.lock = threading.RLock()
        self.logger = Arjuna.get_logger()

        self.__thread_states = {}
        self.__current_issue_counter = 0

    @sync_method('lock')
    def get_current_thread_state(self):
        return self.__thread_states.get(thread_utils.get_current_thread_name())

    @sync_method('lock')
    def get_thread_state(self, tname):
        return self.__thread_states.get(tname)

    @sync_method('lock')
    def register_thread(self, name):
        self.__thread_states[name] = _ThreadState()

    @sync_method('lock')
    def deregister_thread(self, t_name):
        del self.__thread_states[t_name]

    def __get_issue_id(self):
        self.__current_issue_counter += 1
        return self.__current_issue_counter

    @sync_method('lock')
    def create_issue_from_exception(self, e, strace, i_type):
        issue_id = self.__get_issue_id()
        issue = Issue(
            iid = issue_id,
            itype = i_type,
            ename = e.__class__.__name__,
            emsg = str(e),
            etrace = strace
        )

        return issue

    @sync_method('lock')
    def create_issue_for_problem_step(self, i_type):
        issue_id = self.__get_issue_id()
        issue = Issue(
            iid = issue_id,
            itype = i_type,
            ename = "StepException",
            emsg = "Step {}. Refer Step info.".format(i_type.name.lower()),
            etrace = "-"
        )

        return issue


from arjuna.tpi.enums import ArjunaOption

from .guiautomator import GuiAutomator
from .handler import Handler

class Browser(Handler):

    def __init__(self, automator: GuiAutomator):
        super().__init__(automator)
        from arjuna.setuext.guiauto.impl.element.frame import DomRoot
        self.__dom_root = DomRoot(automator)

    @property
    def dom_root(self):
        return self.__dom_root

    def go_to_url(self, url):
        self.automator.dispatcher.go_to_url(url=url)

    def go_back(self, url):
        self.automator.dispatcher.go_back()

    def go_forward(self, url):
        self.automator.dispatcher.go_forward()

    def refresh(self, url):
        self.automator.dispatcher.refresh()

    def execute_javascript(self, js):
        self.automator.dispatcher.execute_javascript(script=js)


from selenium.webdriver.common.by import By

from .melement import MultiElement

class ElementFinder:
    BY_MAP = {
        "ID": By.ID,
        "NAME": By.NAME,
        "TAG_NAME": By.TAG_NAME,
        "CLASS_NAME": By.CLASS_NAME,
        "LINK_TEXT": By.LINK_TEXT,
        "PARTIAL_LINK_TEXT": By.PARTIAL_LINK_TEXT,
        "CSS_SELECTOR": By.CSS_SELECTOR,
        "XPATH": By.XPATH
    }
    
    @classmethod
    def find_element(cls, container, byType, byValue):
        return container.find_element(cls.BY_MAP[byType.upper()], byValue)

    @classmethod
    def find_elements(cls, container, byType, byValue):
        return MultiElement(container.find_elements(cls.BY_MAP[byType.upper()], byValue))



from .handler import Handler

class WebAlertHandler(Handler):

    def __init__(self, automator):
        super().__init__(automator)
        self.__alert_present = False
        self.__alert_setu_id = None
        self.__alert = None

    def create_alert(self):
        if self.__alert_present:
            return self.__alert
        else:
            self.wait()
            from arjuna.setuext.guiauto.impl.element.webalert import WebAlert
            self.__alert_present = True
            alert = WebAlert(self.automator)
            self.__alert_setu_id = alert.setu_id
            self.__alert = alert
            return alert

    def get_alert_for_setu_id(self, setu_id):
        msg = "The alert represented by this object no longer exists."
        if not self.__alert_present:
            raise Exception(msg)
        elif self.__alert_setu_id != setu_id:
            raise Exception(msg)
        else:
            return self.__alert

    def delete_alert(self):
        self.__alert_present = False
        self.__alert_setu_id = None
        self.__alert = None

    def wait(self):
        self.automator.conditions.AlertIsPresent().wait()

    def is_alert_present(self):
        return self.automator.dispatcher.is_web_alert_present()






from flask import Flask
from flask_restful import Api
from waitress import serve

import time

def __launch_setu_svc(port):
    from arjuna.setu.interface.setu import SetuSvc

    from arjuna.setu.config.processor import ConfigCreator
    ConfigCreator.init()

    app = Flask(__name__)
    api = Api(app)

    api.add_resource(SetuSvc, '/setu', endpoint='setu')

    # api.add_resource(ItemList, '/items', endpoint='items')
    #app.run(port=port, use_evalex=False) #, debug=True)
    serve(app, host="localhost", port=port, _quiet=True)

def wait_for_port(port):
    import socket
    server_address = ('localhost', port)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    ct = time.time()
    while(time.time() - ct < 60):
        try:
            sock.bind(server_address)
            sock.close()
            return
        except Exception as e:
            print("here", e)
            time.sleep(1)
    print("Port is not open. Timeout after 60 seconds.")
    raise RuntimeError("SET_SVC_ERROR:: Another service is running at port {}. Setu could not be launched. Message: ".format(port))

def launch_setu(port):
    from arjuna.client.core.connector import _DefaultSetuRequester, _DefaultSetuRequest 
    from arjuna.client.core.action import SetuActionType
    from arjuna.client.core.config import ArjunaComponent
    requester = _DefaultSetuRequester()
    request = _DefaultSetuRequest(ArjunaComponent.SETU, SetuActionType.HELLO)
    response = None
    # try:
    #     response = requester.post(request)
    # except Exception as e:
    #     print("Waiting for port to be free")
    #     wait_for_port(port)
    #     print("Port is open")
    #     __launch_setu_svc(port)  
    #     print("Arjuna Setu service launched.") 
    # else:
    #     rstring = response.get_value().as_string()
    #     if rstring.lower().find("hello") == -1:
    #         raise RuntimeError("SET_SVC_ERROR:: Unexpected Setu error for Hello. Got response: " + rstring)
    #     else:
    #         print("Arjuna Setu service is already running.")

    try:
        wait_for_port(port)
        __launch_setu_svc(port)          
    except Exception as e:
        raise RuntimeError("SET_SVC_ERROR:: Not able to launch Setu Service. Got response: ", e)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import traceback

from .base import TestObject
from arjuna.unitee import Unitee
from arjuna.unitee.reporter.result.types import *
from arjuna.unitee.enums import *
from arjuna.unitee.exceptions import *


class Test(TestObject):

	def __init__(self, num, group, module, func):
		super().__init__(TestObjectTypeEnum.Test)
		self.num = num
		self.group = group
		self.module = module
		self.func = func
		self.func_obj = func.func_obj
		self.append_before_fixture(self.module.fixture_defs.build(FixtureTypeEnum.INIT_EACH_TEST))
		self.append_after_fixture(self.module.fixture_defs.build(FixtureTypeEnum.END_EACH_TEST))

	def load(self, base_tvars):
		self._populate_tvars(base_tvars)

	def _populate_tvars(self, base_tvars):
		self.tvars = base_tvars
		self.tvars.info.test_num = self.num
		self.tvars.info.object_type = TestObjectTypeEnum.Test.name
		# print("while setting", self.num, self.tvars.data.record)

	def _execute(self):

		self.unitee.state_mgr.get_current_thread_state().begin_recording(ExecutionContext.Test)
		try:
			utvars = self.tvars.create_utvars()
			# print(self.func_obj)
			self.func_obj(utvars)
			# print("source", self.num, self.tvars.data.record)
			# print("dest", self.num, utvars.data.record)
		except StepResultEvent as e:
			pass
			# Step addition would have happended as a result of step.evaluate()
		except Exception as e:
			self.unitee.state_mgr.get_current_thread_state().add_step_exception(e, traceback.format_exc())

		steps = self.unitee.state_mgr.get_current_thread_state().end_recording()
		tresult = TestResult(self)
		tresult.set_steps(steps)

		test_reported = False
		if tresult.has_issues():
			test_reported = True
			self.report(tresult)

		for after_fixture in self.after_fixtures:
			fresult = after_fixture.execute(self, self.tvars.clone())
			if fresult.has_issues():
				after_fixture.report()
				if not test_reported:
					tresult.rtype = fresult.rtype
					tresult.rcode = ResultCodeEnum["{}_{}".format(fresult.rtype, after_fixture.type.name)].name
					tresult.iid = fresult.iid
					self.report(tresult)
					test_reported = True
			else:
				if not test_reported:
					if tresult.rtype == "-":
						tresult.rtype = ResultTypeEnum.PASS.name
						tresult.rcode = ResultCodeEnum.PASS_ALL_STEPS.name
					self.report(tresult)
					test_reported = True
					after_fixture.report()

		if not test_reported:
			if tresult.rtype == "-":
				tresult.rtype = ResultTypeEnum.PASS.name
				tresult.rcode = ResultCodeEnum.PASS_ALL_STEPS.name
			self.report(tresult)

	def report(self, result):
		super().report(result)
		self.func.state.update(result)
		self.func.state.test_summary.increment(ResultTypeEnum[result.rtype])


import uuid
from arjuna.setuext.guiauto.impl.automator.guiautomator import GuiAutomator
from arjuna.setuext.guiauto.impl.locator.emd import GuiElementMetaData
from arjuna.setu.adapter import Handler
from arjuna.client.core.action import *
from .action_map import *

# Arg names of methods show JSON names, so don't follow Python conventions.
class GuiAutomatorHandler(Handler):

    def __init__(self, dispatcher_creator):
        self.__dispatcher_creator = dispatcher_creator
        self.__automator = None

    def launch_automator(self, config, extendedConfig=None):
        self.__automator = GuiAutomator(config, extendedConfig)
        self.__automator.dispatcher_creator = self.__dispatcher_creator
        self.__automator.launch()

    def take_action(self, action_type, json_args):
        action_id = HANDLER_MAP[action_type]
        method_name, replaceable = HANDLER_NAME_MAP[action_id]
        return getattr(self, method_name)(action_type.name.replace(replaceable, "").lower(), json_args)

    def take_direct_action(self, action, json_args):
        return getattr(self, action)(**json_args)

    @property
    def setu_id(self):
        return self.automator.setu_id

    @property
    def automator(self):
        return self.__automator

    def quit_automator(self):
        self.automator.quit()

    def define_element(self, locators):
        elem = self.automator.create_element(GuiElementMetaData.createEMD(locators))
        return {"elementSetuId" : elem.setu_id}

    def define_element_with_emd(self, emd):
        elem = self.automator.create_element(emd)
        return {"elementSetuId" : elem.setu_id}

    def define_multielement(self, locators):
        elem = self.automator.create_multielement(GuiElementMetaData.createEMD(locators))
        return {"elementSetuId" : elem.setu_id}

    def define_multielement_with_emd(self, emd):
        elem = self.automator.create_multielement(emd)
        return {"elementSetuId" : elem.setu_id}

    def define_dropdown(self, locators):
        dropdown = self.automator.create_dropdown(GuiElementMetaData.createEMD(locators))
        return {"elementSetuId" : dropdown.setu_id}

    def define_dropdown_with_emd(self, emd):
        dropdown = self.automator.create_dropdown(emd)
        return {"elementSetuId" : dropdown.setu_id}

    def define_radiogroup(self, locators):
        radiogroup = self.automator.create_radiogroup(GuiElementMetaData.createEMD(locators))
        return {"elementSetuId" : radiogroup.setu_id}

    def define_radiogroup_with_emd(self, emd):
        radiogroup = self.automator.create_radiogroup(emd)
        return {"elementSetuId" : radiogroup.setu_id}

    def define_alert(self):
        alert = self.automator.alert_handler.create_alert()
        return {"elementSetuId" : alert.setu_id}

    def define_main_window(self):
        return {"elementSetuId" : self.automator.main_window.setu_id}

    def set_slomo(self, on, interval=None):
        self.automator.set_slomo(on, None)

    def take_window_action(self, action, json_dict):
        elem_setu_id = self.get_element_setuid(json_dict)
        win =  self.automator.main_window.get_window_for_setu_id(elem_setu_id)
        return getattr(WindowHandler, action)(win, **json_dict)

    def take_main_window_action(self, action, json_dict):
        elem_setu_id = self.get_element_setuid(json_dict)
        self.automator.slomo()
        win =  self.automator.main_window.get_window_for_setu_id(elem_setu_id)
        if not win.is_main_window():
            raise Exception("Window represented by {} is not the main window.".format(elem_setu_id))
        return getattr(MainWindowHandler, action)(self.automator.main_window, **json_dict)

    def take_child_window_action(self, action, json_dict):
        elem_setu_id = self.get_element_setuid(json_dict)
        self.automator.slomo()
        win =  self.automator.main_window.get_window_for_setu_id(elem_setu_id)
        return getattr(ChildWindowHandler, action)(win, **json_dict)

    def take_browser_action(self, action, json_dict):
        self.automator.slomo()
        return getattr(BrowserHandler, action)(self.automator.browser, **json_dict)

    def take_element_action(self, action, json_dict):
        elem_setu_id = self.get_element_setuid(json_dict)
        self.automator.slomo()
        instance_action = False
        if "isInstanceAction" in json_dict:
            instance_action = json_dict["isInstanceAction"]
            del json_dict["isInstanceAction"]
        if instance_action:
            index = json_dict["instanceIndex"]
            del json_dict["instanceIndex"]
            multi_element =  self.automator.get_multielement_for_setu_id(elem_setu_id)
            element = multi_element.get_instance_at_index(index)
        else:
            element =  self.automator.get_element_for_setu_id(elem_setu_id)
        return getattr(ElementHandler, action)(element, **json_dict)

    def take_multielement_action(self, action, json_dict):
        elem_setu_id = self.get_element_setuid(json_dict)
        multi_element =  self.automator.get_multielement_for_setu_id(elem_setu_id)
        return getattr(MultiElementHandler, action)(multi_element, **json_dict)

    def take_dropdown_action(self, action, json_dict):
        elem_setu_id = self.get_element_setuid(json_dict)
        self.automator.slomo()
        dropdown =  self.automator.get_element_for_setu_id(elem_setu_id)
        return getattr(DropdownHandler, action)(dropdown, **json_dict)

    def take_radiogroup_action(self, action, json_dict):
        elem_setu_id = self.get_element_setuid(json_dict)
        self.automator.slomo()
        radiogroup =  self.automator.get_element_for_setu_id(elem_setu_id)
        return getattr(RadioGroupHandler, action)(radiogroup, **json_dict)

    def take_alert_action(self, action, json_dict):
        elem_setu_id = self.get_element_setuid(json_dict)
        self.automator.slomo()
        alert =  self.automator.alert_handler.get_alert_for_setu_id(elem_setu_id)
        return getattr(AlertHandler, action)(alert, **json_dict)

    def take_domroot_action(self, action, json_dict):
        return getattr(DomRootHandler, action)(self.automator.browser.dom_root, **json_dict) 

    def create_frame_with_emd(self, emd):
        return self.take_domroot_action("create_frame_with_emd", {'emd':emd})

    def take_frame_action(self, action, json_dict):
        elem_setu_id = self.get_element_setuid(json_dict)
        frame =  self.automator.get_element_for_setu_id(elem_setu_id)
        return getattr(FrameHandler, action)(frame, **json_dict)


# Separates the underlying structure and names
# Also builds json response data where applicable
class ElementHandler:

    @classmethod
    def enter_text(cls, element, text):
        element.enter_text(text)

    @classmethod
    def set_text(cls, element, text):
        element.set_text(text)

    @classmethod
    def click(cls, element):
        element.click()

    @classmethod
    def check(cls, element):
        element.check()

    @classmethod
    def uncheck(cls, element):
        element.uncheck()

    @classmethod
    def identify(cls, element):
        element.identify()

    @classmethod
    def wait_until_present(cls, element):
        element.wait_until_present()

    @classmethod
    def wait_until_visible(cls, element):
        element.wait_until_visible()

    @classmethod
    def wait_until_clickable(cls, element):
        element.wait_until_clickable()

    @classmethod
    def get_source(cls, element):
        return {"value" : element.get_source()}

class MultiElementHandler:

    @classmethod
    def get_instance_count(cls, element):
        return {"value" : element.get_instance_count()}

    @classmethod
    def get_random_index(cls, element):
        return {"value" : element.get_random_index()}

class DropdownHandler:

    @classmethod
    def has_visible_text_selected(cls, dropdown, text):
        return {"checkResult" : dropdown.has_visible_text_selected(text)}

    @classmethod
    def has_value_selected(cls, dropdown, value):
        return {"checkResult" : dropdown.has_value_selected(value)}

    @classmethod
    def has_index_selected(cls, dropdown, index):
        return {"checkResult" : dropdown.has_index_selected(index)}

    @classmethod
    def get_first_selected_option_text(cls, dropdown):
        return {"text" : dropdown.get_first_selected_option_text()}

    @classmethod
    def select_by_visible_text(cls, dropdown, text):
        return dropdown.select_by_visible_text(text)

    @classmethod
    def select_by_value(cls, dropdown, value):
        return dropdown.select_by_value(value)

    @classmethod
    def select_by_index(cls, dropdown, index):
        return dropdown.select_by_index(index)

class RadioGroupHandler:

    @classmethod
    def has_visible_text_selected(cls, radiogroup, text):
        return {"checkResult" : radiogroup.has_visible_text_selected(text)}

    @classmethod
    def has_value_selected(cls, radiogroup, value):
        return {"checkResult" : radiogroup.has_value_selected(value)}

    @classmethod
    def has_index_selected(cls, radiogroup, index):
        return {"checkResult" : radiogroup.has_index_selected(index)}

    @classmethod
    def get_first_selected_option_value(cls, radiogroup):
        return {"value" : radiogroup.get_first_selected_option_value()}

    @classmethod
    def select_by_value(cls, radiogroup, value):
        return radiogroup.select_by_value(value)

    @classmethod
    def select_by_index(cls, radiogroup, index):
        return radiogroup.select_by_index(index)

class WindowHandler:

    @classmethod
    def focus(cls, window):
        return window.focus()

    @classmethod
    def get_title(cls, window):
        return {"title" : window.get_title()}

class MainWindowHandler:

    @classmethod
    def maximize(cls, window):
        return window.maximize()

    @classmethod
    def close_all_child_windows(self, window):
        window.close_all_child_windows() 

    @classmethod
    def get_latest_child_window(self, window):
        return {"elementSetuId" : window.get_latest_child_window().setu_id}

class ChildWindowHandler:

    @classmethod
    def close(cls, window):
        return window.close()

class AlertHandler:

    @classmethod
    def confirm(cls, alert):
        alert.confirm()

    @classmethod
    def dismiss(cls, alert):
        alert.dismiss()

    @classmethod
    def get_text(cls, alert):
        return {"text" : alert.get_text()}

    @classmethod
    def send_text(cls, alert, text):
        alert.send_text(text)

class BrowserHandler:

    @classmethod
    def go_to_url(cls, browser, url):
       browser.go_to_url(url)

    @classmethod
    def go_back(cls, browser):
        browser.go_back()

    @classmethod
    def go_forward(cls, browser):
        browser.go_forward()

    @classmethod
    def refersh(cls, browser):
        browser.refersh()

    @classmethod
    def execute_javascript(self, browser, script):
        browser.execute_javascript(script) 

class DomRootHandler:

    @classmethod
    def focus(cls, dom_root):
        return dom_root.focus()

    @classmethod
    def create_frame(cls, dom_root, locators):
        return {"elementSetuId" : dom_root.create_frame(GuiElementMetaData.createEMD(locators)).setu_id}

    @classmethod
    def create_frame_with_emd(self, dom_root, emd):
        return {"elementSetuId" : dom_root.create_frame(emd).setu_id}

class FrameHandler:

    @classmethod
    def focus(cls, frame):
        return frame.focus()

    @classmethod
    def get_parent(cls, frame):
        return {"elementSetuId" : frame.get_parent().setu_id}

    @classmethod
    def create_frame(cls, frame, locators):
        return {"elementSetuId" : frame.create_frame(GuiElementMetaData.createEMD(locators)).setu_id}


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import abc
import threading
import types

from arjuna.lib.reader.excel import *
from arjuna.lib.reader.ini import *
from arjuna.lib.reader.textfile import *
from arjuna.lib.thread import decorators
from .exceptions import *
from arjuna.setu.types import SetuManagedObject
from arjuna.lib.types import constants
# from arjuna.lib.utils import sys_utils


class ListDataRecord:

    def __init__(self, data_record):
        self.__data_record = data_record

    @property
    def record(self):
        return self.__data_record


class MapDataRecord:

    def __init__(self, data_record):
        self.__data_record = {i.lower():j for i,j in dict(data_record).items()}

    @property
    def record(self):
        return self.__data_record

    def should_exclude(self):
        if "exclude" not in self.__data_record:
            return False

        exclude = self.__data_record['exclude']
        if exclude.upper() in constants.TRUES:
            return True
        else:
            del self.__data_record["exclude"]
            return False


class DataSource(SetuManagedObject, metaclass=abc.ABCMeta):
    def __init__(self):
        super().__init__()
        self.lock = threading.RLock()
        self.name = None
        self.ended = False

    def get_name(self):
        return self.name

    def set_name(self, name):
        self.name = name

    @decorators.sync_method('lock')
    def terminate(self):
        self.ended = True

    def is_terminated(self):
        return self.ended

    def validate(self):
        pass

    @decorators.sync_method('lock')
    def next(self):
        if self.is_terminated():
            raise DataSourceFinished()

        try:
            data_record = self.get_next()
            if not data_record:
                raise StopIteration()
            record = self.process(data_record)
            if self.should_exclude(record):
                return self.next()
            else:
                return record
        except StopIteration:
            raise DataSourceFinished("Records Finished.")
        except Exception as e:
            import traceback
            traceback.print_exc()
            raise DataSourceFinished(
                "Problem happened in getting next record. No further records would be provided.")

    def process(self, data_record):
        return data_record

    @abc.abstractmethod
    def get_next(self):
        pass

    @abc.abstractmethod
    def should_exclude(self, data_record):
        return False

    @abc.abstractmethod
    def reset(self):
        pass


class FileDataSource(DataSource, metaclass=abc.ABCMeta):
    def __init__(self, path):
        super().__init__()
        self.__path = path
        self.__reader = None

    @property
    def path(self):
        return self.__path

    @property
    def reader(self):
        return self.__reader

    @reader.setter
    def reader(self, reader):
        self.__reader = reader

    def get_next(self):
        return self.reader.next()

    def should_exclude(self, data_record):
        return False

    def reset(self):
        self._load_file()

    @abc.abstractmethod
    def _load_file(self):
        pass


class DsvFileListDataSource(FileDataSource):
    def __init__(self, path, delimiter="\t"):
        super().__init__(path)
        self.__delimiter = delimiter
        self._load_file()

    def process(self, data_record):
        return ListDataRecord(data_record)

    def _load_file(self):
        self.reader = FileLine2ArrayReader(self.path, self.__delimiter)


class DsvFileMapDataSource(FileDataSource):
    def __init__(self, path, delimiter="\t"):
        super().__init__(path)
        self.__delimiter = delimiter
        self._load_file()

    def process(self, data_record):
        return MapDataRecord(data_record)

    def _load_file(self):
        self.reader = FileLine2MapReader(self.path, self.__delimiter)


class IniFileDataSource(FileDataSource):
    def __init__(self, path):
        super().__init__(path)
        if path.lower().endswith("ini"):
            self._load_file()
        else:
            raise Exception("Unsupported file extension for Ini reading.")

    def process(self, data_record):
        return MapDataRecord(data_record)

    def _load_file(self):
        self.reader = IniFile2MapReader(self.path)


class ExcelFileListDataSource(FileDataSource):
    def __init__(self, path):
        super().__init__(path)
        if path.lower().endswith("xls"):
            self._load_file()
        else:
            raise Exception("Unsupported file extension for Excel reading.")

    def _load_file(self):
        self.reader = ExcelRow2ArrayReader(self.path)

    def should_exclude(self, data_record):
        return False

    def process(self, data_record):
        return ListDataRecord(data_record)


class ExcelFileMapDataSource(FileDataSource):
    def __init__(self, path):
        super().__init__(path)
        if path.lower().endswith("xls"):
            self._load_file()
        else:
            raise Exception("Unsupported file extension for Excel reading.")

    def _load_file(self):
        self.reader = ExcelRow2MapReader(self.path)

    def should_exclude(self, data_record):
        return data_record.should_exclude()

    def process(self, data_record):
        return MapDataRecord(data_record)